/* $Id: treohelper.c,v 1.266 2004/12/05 21:42:11 chrisf Exp $ */

/*
TreoHelper - Utility to manage power off of Treo
Copyright (C) 2003 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <PalmOS.h>
#include <System/NetMgr.h>
#include <System/SysEvtMgr.h>
#include <Handspring/HsExt.h>
#include <Handspring/HsAppLaunchCmd.h>
#include <Handspring/NetMasterLibrary.h>
#include <PhoneLib/PhoneGlobals.h>
#include <PhoneLib/SMSLibrary.h>
#include "treohelperRsc.h"
#include "treohelper.h"

/* globals */

Err errno, h_errno;
Int32 AppNetTimeout;
UInt16 AppNetRefnum;
UInt16 tps;
UInt16 phonelibref;
struct st300toneid {
    UInt8 slot;
    char dunno[2];
};
void *ToneIDList; /* this is UInt32[] or st300toneid[] */
UInt16 ToneIDCount;

/* prefs */
struct sprefs prefs;

/*
 * This is HORRIBLE!  Part of the PhoneLib is redefined for the 300, with
 * overlapping API and different traps.  I don't see how to get the ringtone
 * API to work smoothly between the GSM and CDMA.  So I'm making functions
 * which call the appropriate trap.
 */

#define Treo300PhnLibTrapGetToneIDs  (sysLibTrapCustom + 42)
#define Treo300PhnLibTrapGetToneName (sysLibTrapCustom + 43)
#define Treo300PhnLibTrapPlayTone    (sysLibTrapCustom + 44)

extern Err Treo300PhnLibGetToneIDs(UInt16 refNum, UInt8 audioType,
    void **list, UInt16* listLength)
    PHN_LIB_TRAP (Treo300PhnLibTrapGetToneIDs);
extern Err Treo300PhnLibGetToneName(UInt16 refNum, UInt16 toneidx, char *name,
    UInt16 maxLength)
    PHN_LIB_TRAP (Treo300PhnLibTrapGetToneName);
extern Err Treo300PhnLibPlayTone(UInt16 refNum, int toneslot, int volume,
    int vibrateState, UInt8 soundType, UInt16 duration,
    Boolean useSoundOffFlag)
    PHN_LIB_TRAP (Treo300PhnLibTrapPlayTone);

/*
 * I have to define this because I am stuck using the old PhoneLib SDK due
 * to the tone library incompatibility.  With the Treo 650 they obsoleted
 * the old method of loading the phone lib.
 */
#define hsSelGetPhoneLibrary 0x86
extern Err HsGetPhoneLibrary(UInt16 * refNum)
    SYS_SEL_TRAP (sysTrapHsSelector, hsSelGetPhoneLibrary);

Boolean MainFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    UInt32 hsStatusVersion;
    static Int32 oldticks=0;
    Int32 ticks;
    ControlPtr ctl;
    ListPtr list;
    Int16 i;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        /* treo gadgets */
        if (FtrGet(hsFtrCreator, hsFtrIDStatusGadgetRev, &hsStatusVersion)
            == 0) {
            HsStatusSetGadgetType(frm, gadgetID_mainhsbattery,
                hsStatusGadgetBattery);
            HsStatusSetGadgetType(frm, gadgetID_mainhssignal,
                hsStatusGadgetSignal);
        }
        UpdateLabels();
        UpdateValues();

        /* set category to 0 (General) */
        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_maincategory));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_maincategory));
        CtlSetLabel(ctl, LstGetSelectionText(list, 0));
        LstSetSelection(list, 0);

        /* set initial UI values from prefs */

        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkID_mainenablephoneonreset)),
            prefs.phoneflags&PHONEFLAGS_ENABLEPHONEONRESET);

        switch (prefs.phoneonofftype) {
        case 0: /* daily */
            CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_maindaily)), 1);
            CtlHitControl(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_maindaily)));
            break;
        case 1: /* weekly */
            CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_mainweekly)), 1);
            CtlHitControl(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                pbuttonID_mainweekly)));
            break;
        }

        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkID_mainstayonlidnetwork)),
            prefs.phoneflags&PHONEFLAGS_STAYONLIDNETWORK);

        set_closest_seconds(frm, listID_mainseconds, poptrigID_mainlidofftime,
            prefs.lidofftime);

        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_mainphoneturnon));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_mainphoneturnon));
        i = (prefs.phoneturnon == 0xffff) ? 0 : (prefs.phoneturnon/15)+1;
        CtlSetLabel(ctl, LstGetSelectionText(list, i));
        LstSetSelection(list, i);

        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_mainphoneturnoff));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_mainphoneturnoff));
        i = (prefs.phoneturnoff == 0xffff) ? 0 : (prefs.phoneturnoff/15)+1;
        CtlSetLabel(ctl, LstGetSelectionText(list, i));
        LstSetSelection(list, i);

        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_mainphoneontries));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_mainphoneontries));
        i = prefs.phoneontries-1; /* list starts with 1 */
        CtlSetLabel(ctl, LstGetSelectionText(list, i));
        LstSetSelection(list, i);

        set_closest_minutes(frm, listID_mainminutes,
            poptrigID_mainphoneofftime, prefs.phoneofftime);

        set_closest_minutes(frm, listID_mainminutes,
            poptrigID_mainphoneonevery, prefs.phoneonevery);

        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkID_mainautookconnect)),
            prefs.phoneflags&PHONEFLAGS_AUTOOKCONNECT);

        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkID_mainautologin)),
            prefs.phoneflags&PHONEFLAGS_AUTOLOGIN);

        handled = true;
        break;

    case frmUpdateEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        UpdateLabels();
        UpdateValues();
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        }
       handled = true;
       break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case checkID_mainenablephoneonreset:
            frm = FrmGetActiveForm();
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_mainenablephoneonreset)))) {
                prefs.phoneflags |= PHONEFLAGS_ENABLEPHONEONRESET;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_ENABLEPHONEONRESET;
            }
            SavePrefs(&prefs);
            break;
        case checkID_mainstayonlidnetwork:
            frm = FrmGetActiveForm();
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_mainstayonlidnetwork)))) {
                prefs.phoneflags |= PHONEFLAGS_STAYONLIDNETWORK;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_STAYONLIDNETWORK;
            }
            SavePrefs(&prefs);
            break;
        case checkID_mainautookconnect:
            frm = FrmGetActiveForm();
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_mainautookconnect)))) {
                prefs.phoneflags |= PHONEFLAGS_AUTOOKCONNECT;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_AUTOOKCONNECT;
            }
            SavePrefs(&prefs);
            /* requires that I enable/disable a service class */
            phone_events_register(phonelibref, &prefs);
            break;
        case checkID_mainautologin:
            frm = FrmGetActiveForm();
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_mainautologin)))) {
                prefs.phoneflags |= PHONEFLAGS_AUTOLOGIN;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_AUTOLOGIN;
            }
            SavePrefs(&prefs);
            sync_autologin_setting(&prefs, true);
            break;
        case poptrigID_mainphoneofftime:
        case poptrigID_mainphoneonevery:
            /*
             * These all share the minuteslist and so I'll just init
             * the list to no selection when it is popped up.  Otherwise
             * it will have the last selection made from any of the
             * triggers and not just the current one.
             */
            frm = FrmGetActiveForm();
            list = FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, listID_mainminutes));
            LstSetSelection(list, -1);
            break;
        case pbuttonID_maindaily:
            prefs.phoneonofftype = 0; /* daily */
            SavePrefs(&prefs);
            /* hide/unhide the appropriate UI */
            frm = FrmGetActiveForm();
            CtlHideControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, buttonID_mainweektimes)));
            CtlShowControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, poptrigID_mainphoneturnon)));
            CtlShowControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, poptrigID_mainphoneturnoff)));
            break;
        case pbuttonID_mainweekly:
            prefs.phoneonofftype = 1; /* weekly */
            SavePrefs(&prefs);
            /* hide/unhide the appropriate UI */
            frm = FrmGetActiveForm();
            CtlHideControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, poptrigID_mainphoneturnon)));
            CtlHideControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, poptrigID_mainphoneturnoff)));
            CtlShowControl(FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, buttonID_mainweektimes)));
            break;
        case buttonID_mainweektimes:
            FrmPopupForm(WeekForm);
            break;
        }
        break;

    case popSelectEvent:
        i = e->data.popSelect.selection;
        if (i < 0) break;
        switch (e->data.ctlSelect.controlID) {
        case poptrigID_mainlidofftime:
            /* sanity check */
            if (i >= SECONDSLIST) i = SECONDSLIST-1;
            prefs.lidofftime = secondslist[i];
            SavePrefs(&prefs);
            break;
        case poptrigID_mainphoneofftime:
            /* sanity check */
            if (i >= MINUTESLIST) i = MINUTESLIST-1;
            prefs.phoneofftime = minuteslist[i];
            SavePrefs(&prefs);
            /* cancel any current phoneoff alarm, if any */
            FtrSet(MainAppID, FTR_COVERAGELOSS, 0);
            set_next_alarm_ref();
            break;
        case poptrigID_mainphoneturnon:
            prefs.phoneturnon = (i <= 0) ? 0xffff : (i-1)*15;
            SavePrefs(&prefs);
            set_phoneonoff_times(&prefs, 0);
            break;
        case poptrigID_mainphoneturnoff:
            prefs.phoneturnoff = (i <= 0) ? 0xffff : (i-1)*15;
            SavePrefs(&prefs);
            set_phoneonoff_times(&prefs, 0);
            break;
        case poptrigID_mainphoneontries:
            prefs.phoneontries = i+1; /* list starts with 1 */
            SavePrefs(&prefs);
            break;
        case poptrigID_mainphoneonevery:
            /* sanity check */
            if (i >= MINUTESLIST) i = MINUTESLIST-1;
            prefs.phoneonevery = minuteslist[i];
            SavePrefs(&prefs);
            break;
        case poptrigID_maincategory:
            switch (i) {
            case 1: /* Drop Net */
                FrmGotoForm(DropnetForm);
                break;
            case 2: /* SMS */
                FrmGotoForm(SMSForm);
                break;
            case 3: /* Fun */
                FrmGotoForm(FunForm);
                break;
            }
            break;
        }
        break;

    case nilEvent:
        ticks = TimGetTicks();
        /* every second update the values */
        if ((ticks-oldticks) >= (tps*1)) {
            oldticks = ticks;
            UpdateValues();
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean DropnetFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    UInt32 hsStatusVersion;
    ControlPtr ctl;
    ListPtr list;
    Int16 i;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        /* treo gadgets */
        if (FtrGet(hsFtrCreator, hsFtrIDStatusGadgetRev, &hsStatusVersion)
            == 0) {
            HsStatusSetGadgetType(frm, gadgetID_dropnethsbattery,
                hsStatusGadgetBattery);
            HsStatusSetGadgetType(frm, gadgetID_dropnethssignal,
                hsStatusGadgetSignal);
        }

        /* set category to 1 (Dropnet) */
        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_dropnetcategory));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_dropnetcategory));
        CtlSetLabel(ctl, LstGetSelectionText(list, 1));
        LstSetSelection(list, 1);

        /* set initial UI values from prefs */

        set_closest_seconds2(frm, listID_dropnetseconds2,
            poptrigID_dropnetgreentime, prefs.dropnetgreentime);

        set_closest_seconds2(frm, listID_dropnetseconds2,
            poptrigID_dropnetgreytime, prefs.dropnetgreytime);

        set_closest_seconds2(frm, listID_dropnetseconds2,
            poptrigID_dropnetcoverage, prefs.dropnetcoverage);

        handled = true;
        break;

    case frmUpdateEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        }
       handled = true;
       break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case poptrigID_dropnetgreentime:
        case poptrigID_dropnetgreytime:
        case poptrigID_dropnetcoverage:
            /*
             * These all share the secondslist and so I'll just init
             * the list to no selection when it is popped up.  Otherwise
             * it will have the last selection made from any of the
             * triggers and not just the current one.
             */
            frm = FrmGetActiveForm();
            list = FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, listID_dropnetseconds2));
            LstSetSelection(list, -1);
            break;
        }
        break;

    case popSelectEvent:
        i = e->data.popSelect.selection;
        if (i < 0) break;
        switch (e->data.ctlSelect.controlID) {
        case poptrigID_dropnetcategory:
            switch (i) {
            case 0: /* General */
                FrmGotoForm(MainForm);
                break;
            case 2: /* SMS */
                FrmGotoForm(SMSForm);
                break;
            case 3: /* Fun */
                FrmGotoForm(FunForm);
                break;
            }
            break;
        case poptrigID_dropnetcoverage:
            /* sanity check */
            if (i >= SECONDSLIST2) i = SECONDSLIST2-1;
            prefs.dropnetcoverage = secondslist2[i];
            SavePrefs(&prefs);
            break;
        case poptrigID_dropnetgreentime:
            /* sanity check */
            if (i >= SECONDSLIST2) i = SECONDSLIST2-1;
            prefs.dropnetgreentime = secondslist2[i];
            SavePrefs(&prefs);
            set_dropnet_times(&prefs, 0);
            break;
        case poptrigID_dropnetgreytime:
            /* sanity check */
            if (i >= SECONDSLIST2) i = SECONDSLIST2-1;
            prefs.dropnetgreytime = secondslist2[i];
            SavePrefs(&prefs);
            set_dropnet_times(&prefs, 0);
            break;
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean SMSFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    UInt32 hsStatusVersion;
    Int16 i;
    ControlPtr ctl;
    ListPtr list;
    FieldPtr fld;
    UInt32 crid;
    char *txt;
    static Int16 which_pbut=0;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        /* treo gadgets */
        if (FtrGet(hsFtrCreator, hsFtrIDStatusGadgetRev, &hsStatusVersion)
            == 0) {
            HsStatusSetGadgetType(frm, gadgetID_smshsbattery,
                hsStatusGadgetBattery);
            HsStatusSetGadgetType(frm, gadgetID_smshssignal,
                hsStatusGadgetSignal);
        }

        /* set category to 2 (SMS) */
        which_pbut = 0;
        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_smscategory));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_smscategory));
        CtlSetLabel(ctl, LstGetSelectionText(list, 2));
        LstSetSelection(list, 2);

        /* set the pushbutton to 1 */
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            pbuttonID_smstrig1)), 1);
        CtlHitControl(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            pbuttonID_smstrig1)));

        handled = true;
        break;

    case frmUpdateEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case frmCloseEvent:
        frm = FrmGetActiveForm();
        /* need to copy UI text into prefs if changed */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_smstrigtext));
        if (FldDirty(fld)) {
            MemSet(prefs.smstrigtext[which_pbut],
                sizeof(prefs.smstrigtext[0]), 0);
            txt = FldGetTextPtr(fld);
            if (txt) MemMove(prefs.smstrigtext[which_pbut], txt, StrLen(txt));
            SavePrefs(&prefs);
        }
        /* need to convert duration text into UInt16 */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_smsduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.smsduration[which_pbut] = (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
        }
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        }
       handled = true;
       break;

    case ctlSelectEvent:
        frm = FrmGetActiveForm();
        /* need to copy UI text into prefs if changed */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_smstrigtext));
        if (FldDirty(fld)) {
            MemSet(prefs.smstrigtext[which_pbut],
                sizeof(prefs.smstrigtext[0]), 0);
            txt = FldGetTextPtr(fld);
            if (txt) MemMove(prefs.smstrigtext[which_pbut], txt, StrLen(txt));
            SavePrefs(&prefs);
            FldSetDirty(fld, false);
        }
        /* need to convert duration text into UInt16 */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_smsduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.smsduration[which_pbut] = (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
            FldSetDirty(fld, false);
        }
        /* pushing the 1-8 changes the active sms trigger */
        if ((e->data.ctlSelect.controlID >= pbuttonID_smstrig1) &&
            (e->data.ctlSelect.controlID <= pbuttonID_smstrigdef)) {
            which_pbut = e->data.ctlSelect.controlID-pbuttonID_smstrig1;
            set_smstrig(frm, which_pbut);
        }
        switch (e->data.ctlSelect.controlID) {
        case checkID_smsenabletrig:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_smsenabletrig)))) {
                prefs.smsflags[which_pbut] |= SMSFLAGS_ENABLETRIG;
            } else {
                prefs.smsflags[which_pbut] &= ~SMSFLAGS_ENABLETRIG;
            }
            SavePrefs(&prefs);
            break;
        case checkID_smsenabletone:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_smsenabletone)))) {
                prefs.smsflags[which_pbut] |= SMSFLAGS_ENABLETONE;
            } else {
                prefs.smsflags[which_pbut] &= ~SMSFLAGS_ENABLETONE;
            }
            SavePrefs(&prefs);
            break;
        case checkID_smsvibrate:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_smsvibrate)))) {
                prefs.smsflags[which_pbut] |= SMSFLAGS_VIBRATE;
            } else {
                prefs.smsflags[which_pbut] &= ~SMSFLAGS_VIBRATE;
            }
            SavePrefs(&prefs);
            break;
        case checkID_smsdelete:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_smsdelete)))) {
                prefs.smsflags[which_pbut] |= SMSFLAGS_DELETE;
            } else {
                prefs.smsflags[which_pbut] &= ~SMSFLAGS_DELETE;
            }
            SavePrefs(&prefs);
            break;
        case checkID_smsdelaylaunch:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_smsdelaylaunch)))) {
                prefs.smsflags[which_pbut] |= SMSFLAGS_DELAYLAUNCH;
            } else {
                prefs.smsflags[which_pbut] &= ~SMSFLAGS_DELAYLAUNCH;
            }
            SavePrefs(&prefs);
            break;
        case poptrigID_smstone:
            prefs.smstone[which_pbut] = ToneList(listID_smstone,
                poptrigID_smstone, prefs.smstone[which_pbut]);
            SavePrefs(&prefs);
            /* demo the tone for the user, no vibrate though */
            TreoPlayTone(phonelibref, prefs.smstone[which_pbut],
                prefs.smsvolume[which_pbut], false,
                prefs.smsduration[which_pbut]);
            break;
        }
        break;

    case popSelectEvent:
        i = e->data.popSelect.selection;
        if (i < 0) break;
        switch (e->data.ctlSelect.controlID) {
        case poptrigID_smscategory:
            switch (i) {
            case 0: /* General */
                FrmGotoForm(MainForm);
                break;
            case 1: /* Drop Net */
                FrmGotoForm(DropnetForm);
                break;
            case 3: /* Fun */
                FrmGotoForm(FunForm);
                break;
            }
            break;
        case poptrigID_smslaunch:
            switch (i) {
            case 1: crid = 'GNGM'; break; /* GNUGotMail */
            case 2: crid = 'PBJT'; break; /* PBJTime */
            case 3: crid = 'Snpm'; break; /* SnapperMail */
            case 4: crid = 'secr'; break; /* Security lock */
            case 5: crid = 'asc3'; break; /* VersaMail */
            default: crid = 0; break;
            }
            prefs.smslaunch[which_pbut] = crid;
            SavePrefs(&prefs);
            //MatchSMStext(phonelibref, &prefs, "testing"); /* testing */
            break;
        case poptrigID_smsvolume:
            prefs.smsvolume[which_pbut] = i;
            SavePrefs(&prefs);
            break;
        }
        break;

    case penDownEvent:
        /* kill any demo tone */
        PhnLibStopTone(phonelibref);
        break;

    default:
        break;
    }

    return handled;
}

Boolean FunFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    UInt32 hsStatusVersion;
    ControlPtr ctl;
    ListPtr list;
    Int16 i;
    FieldPtr fld;
    char *txt;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        /* treo gadgets */
        if (FtrGet(hsFtrCreator, hsFtrIDStatusGadgetRev, &hsStatusVersion)
            == 0) {
            HsStatusSetGadgetType(frm, gadgetID_funhsbattery,
                hsStatusGadgetBattery);
            HsStatusSetGadgetType(frm, gadgetID_funhssignal,
                hsStatusGadgetSignal);
        }

        /* set category to 3 (Fun) */
        list = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_funcategory));
        ctl = FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, poptrigID_funcategory));
        CtlSetLabel(ctl, LstGetSelectionText(list, 3));
        LstSetSelection(list, 3);

        /* set initial UI values from prefs */
        set_fun(frm);

        handled = true;
        break;

    case frmUpdateEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case frmCloseEvent:
        frm = FrmGetActiveForm();
        /* need to copy UI text into prefs if changed */
        /* need to convert durations text into UInt16 */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_funpoweronduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.funduration[0] /* power on */ =
                (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
        }
        /* was a duration changed? */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_funpoweroffduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.funduration[1] /* power off */ =
                (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
        }
        /* was the handsfree dial number changed? */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_funhandsfreenumber));
        if (FldDirty(fld)) {
            MemSet(prefs.funhandsfreenumber,
                sizeof(prefs.funhandsfreenumber), 0);
            txt = FldGetTextPtr(fld);
            if (txt) MemMove(prefs.funhandsfreenumber, txt, StrLen(txt));
            SavePrefs(&prefs);
        }
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        }
       handled = true;
       break;

    case ctlSelectEvent:
        frm = FrmGetActiveForm();
        /* need to convert durations text into UInt16 */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_funpoweronduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.funduration[0] /* power on */ =
                (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
            FldSetDirty(fld, false);
        }
        /* was a duration changed? */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_funpoweroffduration));
        if (FldDirty(fld)) {
            txt = FldGetTextPtr(fld);
            if (txt) prefs.funduration[1] /* power off */ =
                (UInt16)StrAToI(txt);
            SavePrefs(&prefs);
            FldSetDirty(fld, false);
        }
        /* was the handsfree dial number changed? */
        fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_funhandsfreenumber));
        if (FldDirty(fld)) {
            MemSet(prefs.funhandsfreenumber,
                sizeof(prefs.funhandsfreenumber), 0);
            txt = FldGetTextPtr(fld);
            if (txt) MemMove(prefs.funhandsfreenumber, txt, StrLen(txt));
            SavePrefs(&prefs);
            FldSetDirty(fld, false);
        }
        switch (e->data.ctlSelect.controlID) {
        case checkID_funenablepowerontone:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_funenablepowerontone)))) {
                prefs.phoneflags |= PHONEFLAGS_ENABLEPOWERONTONE;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_ENABLEPOWERONTONE;
            }
            SavePrefs(&prefs);
            break;
        case checkID_funenablepowerofftone:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_funenablepowerofftone)))) {
                prefs.phoneflags |= PHONEFLAGS_ENABLEPOWEROFFTONE;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_ENABLEPOWEROFFTONE;
            }
            SavePrefs(&prefs);
            break;
        case poptrigID_funpowerontone:
            prefs.funtone[0] /* power on */ = ToneList(listID_funpowerontone,
                poptrigID_funpowerontone, prefs.funtone[0]);
            SavePrefs(&prefs);
            /* demo the tone for the user */
            TreoPlayTone(phonelibref, prefs.funtone[0],
                prefs.funvolume[0], false,
                prefs.funduration[0]);
            break;
        case poptrigID_funpowerofftone:
            prefs.funtone[1] /* power off */ = ToneList(listID_funpowerofftone,
                poptrigID_funpowerofftone, prefs.funtone[1]);
            SavePrefs(&prefs);
            /* demo the tone for the user */
            TreoPlayTone(phonelibref, prefs.funtone[1],
                prefs.funvolume[1], false,
                prefs.funduration[1]);
            break;
        case checkID_funenablehandsfreedial:
            if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                checkID_funenablehandsfreedial)))) {
                prefs.phoneflags |= PHONEFLAGS_ENABLEHANDSFREEDIAL;
            } else {
                prefs.phoneflags &= ~PHONEFLAGS_ENABLEHANDSFREEDIAL;
            }
            SavePrefs(&prefs);
            /* requires that I enable/disable a service class */
            phone_events_register(phonelibref, &prefs);
            break;
        }
        break;

    case popSelectEvent:
        i = e->data.popSelect.selection;
        if (i < 0) break;
        switch (e->data.ctlSelect.controlID) {
        case poptrigID_funcategory:
            switch (i) {
            case 0: /* General */
                FrmGotoForm(MainForm);
                break;
            case 1: /* Drop Net */
                FrmGotoForm(DropnetForm);
                break;
            case 2: /* SMS */
                FrmGotoForm(SMSForm);
                break;
            }
            break;
        case poptrigID_funpoweronvolume:
            prefs.funvolume[0] /* power on */ = i;
            SavePrefs(&prefs);
            break;
        case poptrigID_funpoweroffvolume:
            prefs.funvolume[1] /* power off */ = i;
            SavePrefs(&prefs);
            break;
        }
        break;

    case penDownEvent:
        /* kill any demo tone */
        PhnLibStopTone(phonelibref);
        break;

    default:
        break;
    }

    return handled;
}

Boolean AboutFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_aboutok:
            FrmReturnToForm(0);
            handled = true;
            break;
        case buttonID_aboutcontrib:
            FrmPopupForm(ContribForm);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean ContribFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    ListPtr lst;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_contrib));
        /* turn off the highlight */
        LstSetSelection(lst, -1);
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_contribok:
            FrmReturnToForm(0);
            handled = true;
            break;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, winUp, LstGetVisibleItems(lst));
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, winDown, LstGetVisibleItems(lst));
        }
        break;

    default:
        break;
    }

    return handled;
}

Boolean WeekFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    Int16 i;
    UInt16 ctlid;
    ListPtr list;

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);

        /* initialize our UI controls */

        /* on times */
        for (i=0; i < 7; i++) {
            set_closest_time(frm, listID_weektimes,
                poptrigID_week0on+i, prefs.phoneweekonoff[i]);
        }
        /* off times */
        for (i=0; i < 7; i++) {
            set_closest_time(frm, listID_weektimes,
                poptrigID_week0off+i, prefs.phoneweekonoff[i+7]);
        }

        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_weekok:
            FrmReturnToForm(0);
            handled = true;
            break;
        case poptrigID_week0on:
        case poptrigID_week1on:
        case poptrigID_week2on:
        case poptrigID_week3on:
        case poptrigID_week4on:
        case poptrigID_week5on:
        case poptrigID_week6on:
        case poptrigID_week0off:
        case poptrigID_week1off:
        case poptrigID_week2off:
        case poptrigID_week3off:
        case poptrigID_week4off:
        case poptrigID_week5off:
        case poptrigID_week6off:
            /*
             * These all share the listID_weektimes and so I'll just init
             * the list to no selection when it is popped up.  Otherwise
             * it will have the last selection made from any of the
             * triggers and not just the current one.
             */
            frm = FrmGetActiveForm();
            list = FrmGetObjectPtr(frm,
                FrmGetObjectIndex(frm, listID_weektimes));
            LstSetSelection(list, -1);
            break;
        }
        break;

    case popSelectEvent:
        i = e->data.popSelect.selection;
        if (i < 0) break;
        ctlid = e->data.ctlSelect.controlID;
        /*
         * Set the appropriate on/off time in our pref array.  Remember that
         * prefs.phoneweekonoff is 7 elements of on times followed by 7
         * elements of off times.  The units is minutes, or 0xffff as N/A.
         */
        if ((ctlid >= poptrigID_week0on) &&
            (ctlid <= poptrigID_week6on)) {
            prefs.phoneweekonoff[ctlid-poptrigID_week0on] =
                (i <= 0) ? 0xffff : (i-1)*15;
        } else if ((ctlid >= poptrigID_week0off) &&
            (ctlid <= poptrigID_week6off)) {
            prefs.phoneweekonoff[ctlid-poptrigID_week0off+7] =
                (i <= 0) ? 0xffff : (i-1)*15;
        }
        /* save and activate new times as we go along */
        SavePrefs(&prefs);
        set_phoneonoff_times(&prefs, 0);
        break;

    default:
        break;
    }

    return handled;
}

void set_closest_seconds(FormPtr frm, UInt16 listID, UInt16 poptrigID,
    UInt16 secs)
{
    ListPtr list;
    ControlPtr ctl;
    UInt16 i;

    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID));
    /* find the closest matching list item to our # of seconds pref */
    for (i=0; i < SECONDSLIST; i++) {
        if (secondslist[i] > secs) { i--; break; }
    }
    /* sanity check */
    if (i < 0) i = 0;
    if (i >= SECONDSLIST) i = SECONDSLIST-1;
    CtlSetLabel(ctl, LstGetSelectionText(list, i));
    LstSetSelection(list, i);
}

void set_closest_seconds2(FormPtr frm, UInt16 listID, UInt16 poptrigID,
    UInt16 secs)
{
    ListPtr list;
    ControlPtr ctl;
    UInt16 i;

    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID));
    /* find the closest matching list item to our # of seconds pref */
    for (i=0; i < SECONDSLIST2; i++) {
        if (secondslist2[i] > secs) { i--; break; }
    }
    /* sanity check */
    if (i < 0) i = 0;
    if (i >= SECONDSLIST2) i = SECONDSLIST2-1;
    CtlSetLabel(ctl, LstGetSelectionText(list, i));
    LstSetSelection(list, i);
}

void set_closest_minutes(FormPtr frm, UInt16 listID, UInt16 poptrigID,
    UInt16 mins)
{
    ListPtr list;
    ControlPtr ctl;
    UInt16 i;

    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID));
    /* find the closest matching list item to our # of minutes pref */
    for (i=0; i < MINUTESLIST; i++) {
        if (minuteslist[i] > mins) { i--; break; }
    }
    /* sanity check */
    if (i < 0) i = 0;
    if (i >= MINUTESLIST) i = MINUTESLIST-1;
    CtlSetLabel(ctl, LstGetSelectionText(list, i));
}

void set_closest_time(FormPtr frm, UInt16 listID, UInt16 poptrigID, UInt16 mins)
{
    ListPtr list;
    ControlPtr ctl;
    UInt16 i;

    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID));
    i = (mins == 0xffff) ? 0 : (mins/15)+1;
    CtlSetLabel(ctl, LstGetSelectionText(list, i));
}

void set_smstrig(FormPtr frm, Int16 which)
{
    FieldPtr fld;
    ListPtr list;
    ControlPtr ctl;
    UInt16 i;
    char *p;
    char txt[6];

    if (which >= SMSENTRIES) return;

    /* enable? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_smsenabletrig)), prefs.smsflags[which]&SMSFLAGS_ENABLETRIG);
    /* message tone? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_smsenabletone)), prefs.smsflags[which]&SMSFLAGS_ENABLETONE);
    /* vibrate? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_smsvibrate)), prefs.smsflags[which]&SMSFLAGS_VIBRATE);
    /* delete? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_smsdelete)), prefs.smsflags[which]&SMSFLAGS_DELETE);
    /* trigger text */
    fld = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, fieldID_smstrigtext));
    FldDelete(fld, 0, FldGetMaxChars(fld));
    FldInsert(fld, prefs.smstrigtext[which], StrLen(prefs.smstrigtext[which]));
    FldSetDirty(fld, false); /* so I know when the user changes it */
    /*
     * message tone - I have stored a toneid (GSM), or a toneslot (CDMA).
     */
    /* need to retrieve tone name and copy it to trigger */
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID_smstone));
    p = (char *)CtlGetLabel(ctl);
    /* the 32 is how many chars we have alloc'd in the trigger */
    TreoGetToneName(phonelibref, prefs.smstone[which], p, 32);
    CtlSetLabel(ctl, p);
    /* launch */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_smslaunch));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID_smslaunch));
    switch (prefs.smslaunch[which]) {
    case 'GNGM': i = 1; break; /* GNUGotMail */
    case 'PBJT': i = 2; break; /* PBJTime */
    case 'Snpm': i = 3; break; /* SnapperMail */
    case 'secr': i = 4; break; /* Security lock */
    case 'asc3': i = 5; break; /* VersaMail */
    default: i = 0; break;
    }
    CtlSetLabel(ctl, LstGetSelectionText(list, i));
    LstSetSelection(list, i);
    /* delay launch? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_smsdelaylaunch)), prefs.smsflags[which]&SMSFLAGS_DELAYLAUNCH);
    /* volume */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_smsvolume));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, poptrigID_smsvolume));
    CtlSetLabel(ctl, LstGetSelectionText(list, prefs.smsvolume[which]));
    LstSetSelection(list, prefs.smsvolume[which]);
    /* duration */
    fld = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, fieldID_smsduration));
    FldDelete(fld, 0, FldGetMaxChars(fld));
    StrPrintF(txt, "%d", prefs.smsduration[which]);
    FldInsert(fld, txt, StrLen(txt));
    FldSetDirty(fld, false); /* so I know when the user changes it */
}

void set_fun(FormPtr frm)
{
    FieldPtr fld;
    ListPtr list;
    ControlPtr ctl;
    char *p;
    char txt[6];

    /* power on tone enable? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_funenablepowerontone)),
        prefs.phoneflags&PHONEFLAGS_ENABLEPOWERONTONE);
    /* power off tone enable? */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_funenablepowerofftone)),
        prefs.phoneflags&PHONEFLAGS_ENABLEPOWEROFFTONE);

    /* power on tone - I have stored a toneid (GSM), or a toneslot (CDMA) */
    /* need to retrieve tone name and copy it to trigger */
    ctl = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, poptrigID_funpowerontone));
    p = (char *)CtlGetLabel(ctl);
    /* the 32 is how many chars we have alloc'd in the trigger */
    TreoGetToneName(phonelibref, prefs.funtone[0] /* power on */, p, 32);
    CtlSetLabel(ctl, p);
    /* power off tone - I have stored a toneid (GSM), or a toneslot (CDMA) */
    /* need to retrieve tone name and copy it to trigger */
    ctl = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, poptrigID_funpowerofftone));
    p = (char *)CtlGetLabel(ctl);
    /* the 32 is how many chars we have alloc'd in the trigger */
    TreoGetToneName(phonelibref, prefs.funtone[1] /* power off */, p, 32);
    CtlSetLabel(ctl, p);

    /* power on volume */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_funpoweronvolume));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        poptrigID_funpoweronvolume));
    CtlSetLabel(ctl, LstGetSelectionText(list,
        prefs.funvolume[0] /* power on */));
    LstSetSelection(list, prefs.funvolume[0]);
    /* power off volume */
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        listID_funpoweroffvolume));
    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        poptrigID_funpoweroffvolume));
    CtlSetLabel(ctl, LstGetSelectionText(list,
        prefs.funvolume[1] /* power off */));
    LstSetSelection(list, prefs.funvolume[1]);

    /* power on duration */
    fld = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, fieldID_funpoweronduration));
    FldDelete(fld, 0, FldGetMaxChars(fld));
    StrPrintF(txt, "%d", prefs.funduration[0] /* power on */);
    FldInsert(fld, txt, StrLen(txt));
    FldSetDirty(fld, false); /* so I know when the user changes it */
    /* power off duration */
    fld = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, fieldID_funpoweroffduration));
    FldDelete(fld, 0, FldGetMaxChars(fld));
    StrPrintF(txt, "%d", prefs.funduration[1] /* power off */);
    FldInsert(fld, txt, StrLen(txt));
    FldSetDirty(fld, false); /* so I know when the user changes it */

    /* enable/disable handsfree button dialing */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        checkID_funenablehandsfreedial)),
        prefs.phoneflags&PHONEFLAGS_ENABLEHANDSFREEDIAL);
    /* number to dial */
    fld = FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, fieldID_funhandsfreenumber));
    FldDelete(fld, 0, FldGetMaxChars(fld));
    FldInsert(fld, prefs.funhandsfreenumber, StrLen(prefs.funhandsfreenumber));
    FldSetDirty(fld, false); /* so I know when the user changes it */
}

void UpdateLabels(void)
{
    FormPtr frm;
    UInt16 x, y;
    RectangleType r1;

    frm = FrmGetActiveForm();
    if ((frm != FrmGetFormPtr(MainForm))) return;
    if (FrmGetWindowHandle(frm) != WinGetDrawWindow()) return;

    FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_mainlabels), &r1);
    x = r1.topLeft.x;
    y = r1.topLeft.y;
    WinEraseRectangle(&r1, 0);
    WinDrawChars("SignalQuality()", 15, x, y);
    y += FntCharHeight();
    WinDrawChars("GetPhoneCallStatus()", 20, x, y);
}

void UpdateValues(void)
{
    char txt[32];
    FormPtr frm;
    UInt16 x, y;
    RectangleType r1;
    UInt16 ui16val;
    UInt32 ui32val;

    frm = FrmGetActiveForm();
    if ((frm != FrmGetFormPtr(MainForm))) return;
    if (FrmGetWindowHandle(frm) != WinGetDrawWindow()) return;

    FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_mainvalues), &r1);
    x = r1.topLeft.x;
    y = r1.topLeft.y;
    WinEraseRectangle(&r1, 0);
    if (!PhnLibSignalQuality(phonelibref, &ui16val)) {
        StrPrintF(txt, "%d", ui16val);
        WinDrawChars(txt, StrLen(txt), x, y);
    }
    y += FntCharHeight();
    if (!PhnLibGetPhoneCallStatus(phonelibref, &ui32val)) {
        StrPrintF(txt, "%08lx", ui32val);
        WinDrawChars(txt, StrLen(txt), x, y);
    }
}

Int16 MatchSMStext(UInt16 plibr, struct sprefs *tprefs, char *txt)
{
    UInt16 i, l;
    Int16 ret=-1;
    void *p;
    UInt16 tplibref;
    UInt32 tui32;
    Boolean phone_on=false;

    /*
     * Determine if we are in a phone call.  If so, do a delayed launch
     * for any match.
     */
    if (!LoadPhoneLibrary(&tplibref)) {
        if (!PhnLibGetPhoneCallStatus(tplibref, &tui32) &&
            (tui32&(phnVoiceCall1Active|phnVoiceCall2Active))) {
            phone_on = true;
        }
        ClosePhoneLibrary(tplibref);
    }

    l = StrLen(txt)+1;
    for (i=0; i < SMSENTRIES; i++) {
        if (tprefs->smsflags[i]&SMSFLAGS_ENABLETRIG) {
            /* see if the trigger text exists within the SMS */
            /* special case for default which doesn't need trigtext */
            if ((i == SMSENTRIES-1) || StrStr(txt, tprefs->smstrigtext[i])) {
                if (phone_on || (tprefs->smsflags[i]&SMSFLAGS_DELAYLAUNCH)) {
                    /* yup, we need to store it in feature memory */
                    /* first get rid of an older one */
                    if (!FtrGet(MainAppID, FTR_SMS1+i, (UInt32 *)&p))
                        FtrPtrFree(MainAppID, FTR_SMS1+i);
                    /* allocate new feature memory for the asciiz */
                    if (!FtrPtrNew(MainAppID, FTR_SMS1+i, l, &p)) {
                        /* write the SMS text into the feature memory */
                        if (DmWrite(p, 0, txt, l)) {
                            /* failed, so free it up */
                            FtrPtrFree(MainAppID, FTR_SMS1+i);
                        }
                    }
                }
                ret = i;
                break; /* i is now which_pbut */
            }
        }
    }

    if (ret == -1) return ret; /* no match */

    /* play the tone we have assigned */
    if (!phone_on && (tprefs->smsflags[i]&SMSFLAGS_ENABLETONE)) {
        TreoPlayTone(plibr, tprefs->smstone[i], tprefs->smsvolume[i],
            tprefs->smsflags[i]&SMSFLAGS_VIBRATE,
            tprefs->smsduration[i]);
    }

    /*
     * Do the launch if it's not delayed.  When we go to sleep, I run through
     * and see if any FTR_SMS are set and I do the launches then.
     */
    if (!phone_on && !(tprefs->smsflags[i]&SMSFLAGS_DELAYLAUNCH))
        do_sms_launch(tprefs, txt, i);

    return ret;
}

void do_sms_launch(struct sprefs *tprefs, char *txt, UInt16 which)
{
    UInt16 i1, i2, l;
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res, retval;
    void *p;

    EvtWakeup(); /* make at least one event is processed */
    l = StrLen(txt)+1;
    /* do whatever launch we have assigned */
    switch (tprefs->smslaunch[which]) {
    case 'GNGM': /* GNUGotMail */
        GGMCheckMail();
        GGMRefreshList();
        EvtSetAutoOffTimer(SetExactly, 10); /* power off in 10 seconds */
        break;
    case 'PBJT': /* PBJTime */
        if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
            sysFileTApplication, 'PBJT', true, &cardNo, &dbID) == 0)) {
            p = MemPtrNew(l);
            if (!p) break;
            /* copy everything but the \r chars */
            for (i1=0,i2=0; i1 < l; i1++) {
                if (txt[i1] != '\r') ((char *)p)[i2++] = txt[i1];
            }
            MemPtrSetOwner(p, 0);
            retval = SysAppLaunch(cardNo, dbID, sysAppLaunchFlagNewStack,
                sysAppLaunchCmdCustomBase+100, p, &res);
            MemPtrFree(p); /* have to free with SysAppLaunch */
        }
        break;
    case 'Snpm': /* SnapperMail */
        if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
            sysFileTApplication, 'Snpm', true, &cardNo, &dbID) == 0)) {
            retval = SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdCustomBase,
                NULL);
        }
        break;
    case 'secr': /* Security lock */
        if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
            sysFileTApplication, 'secr', true, &cardNo, &dbID) == 0)) {
            retval = SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdSystemLock, 0);
        }
        break;
    case 'asc3': /* VersaMail */
        VersaMailCheck();
        break;
    }
}

void VersaMailCheck(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 retval;

    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, 'asc3', true, &cardNo, &dbID) == 0)) {
        retval = SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdCustomBase+3,
            NULL);
    }
}

void GGMCheckMail(void)
{
    UInt16 cardNo;
    LocalID dbID;
    UInt32 res;
    DmSearchStateType searchInfo;
    UInt32 retval;
    FormActiveStateType formState;
    unsigned char *p;

    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, 'GNGM', true, &cardNo, &dbID) == 0)) {
        p = MemPtrNew(sizeof(unsigned char));
        if (!p) return;
        *p = 0; /* Check Mail */
        MemPtrSetOwner(p, 0);
        /* save all of the current app's UI state */
        FrmSaveActiveState(&formState);
        retval = SysAppLaunch(cardNo, dbID,
            sysAppLaunchFlagNewStack|sysAppLaunchFlagNewGlobals,
            sysAppLaunchCmdCustomBase+1, (void *)p, &res);
        MemPtrFree(p); /* have to free with SysAppLaunch */
        /* restore all of the current app's UI state */
        FrmRestoreActiveState(&formState);
    }
}

void GGMRefreshList(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res;
    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, 'mail', true, &cardNo, &dbID) == 0)) {
        SysAppLaunch(cardNo, dbID, 0, sysAppLaunchCmdCustomBase+2,
            NULL, &res);
    }
}

/*
 * Returns field that has the focus, if any, including in embedded tables.
 */
FieldPtr GetFocusObjectPtr(void)
{
    FormPtr frm;
    UInt16 focus;
    FormObjectKind objType;

    frm = FrmGetActiveForm();
    focus = FrmGetFocus(frm);
    if (focus == noFocus) return NULL;
    objType = FrmGetObjectType(frm, focus);
    if (objType == frmFieldObj) return FrmGetObjectPtr(frm, focus);
    else if (objType == frmTableObj)
        return TblGetCurrentField(FrmGetObjectPtr(frm, focus));
    return NULL;
}

Boolean FormHandleMenuEvent(UInt16 menuID)
{
    FieldPtr fld;

    switch (menuID) {
    case sysEditMenuUndoCmd:
    case sysEditMenuCutCmd:
    case sysEditMenuCopyCmd:
    case sysEditMenuPasteCmd:
    case sysEditMenuSelectAllCmd:
        fld = GetFocusObjectPtr();
        if (!fld) return false;
        if (menuID == sysEditMenuUndoCmd) FldUndo(fld);
        else if (menuID == sysEditMenuCutCmd) FldCut(fld);
        else if (menuID == sysEditMenuCopyCmd) FldCopy(fld);
        else if (menuID == sysEditMenuPasteCmd) FldPaste(fld);
        else if (menuID == sysEditMenuSelectAllCmd)
            FldSetSelection (fld, 0, FldGetTextLength (fld));
        return true;
    case sysEditMenuKeyboardCmd:
        SysKeyboardDialog(kbdDefault);
        return true;
    case sysEditMenuGraffitiCmd:
        SysGraffitiReferenceDialog(referenceDefault);
        return true;
    }
    return false;
}

Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    UInt16 formId;
    Boolean handled;

    handled = false;
    if (e->eType == frmLoadEvent) {
        formId = e->data.frmLoad.formID;
        frm = FrmInitForm(formId);
        FrmSetActiveForm(frm);

        switch(formId) {
        case MainForm:
            FrmSetEventHandler(frm, MainFormHandleEvent);
            break;
        case DropnetForm:
            FrmSetEventHandler(frm, DropnetFormHandleEvent);
            break;
        case SMSForm:
            FrmSetEventHandler(frm, SMSFormHandleEvent);
            break;
        case FunForm:
            FrmSetEventHandler(frm, FunFormHandleEvent);
            break;
        case AboutForm:
            FrmSetEventHandler(frm, AboutFormHandleEvent);
            break;
        case ContribForm:
            FrmSetEventHandler(frm, ContribFormHandleEvent);
            break;
        case WeekForm:
            FrmSetEventHandler(frm, WeekFormHandleEvent);
            break;
        }
        handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
UInt16 StartApplication(void)
{
    Err err;
    Boolean inposer=false;
    HostID hid;

    tps = SysTicksPerSecond();
    AppNetTimeout = 60*tps;
    AppNetRefnum = -1;
    ToneIDList = NULL;
    ToneIDCount = 0;
    LoadPrefs(&prefs, WHICHPREFS_ALL);

    hid = HostGetHostID();
    if ((hid == hostIDPalmOSEmulator) || (hid == hostIDPalmOSSimulator))
        inposer = true;

    if (LoadPhoneLibrary(&phonelibref)) {
        FrmCustomAlert(ErrorAlert, "Need Phone Library.", " ", " ");
        return 1;
    }

    if (!PhnLibModulePowered(phonelibref) && !inposer) {
        FrmCustomAlert(ErrorAlert,
            "Phone must be turned on while using this application.",
            " ", " ");
        /* maybe they turned it on before hitting OK */
        if (!PhnLibModulePowered(phonelibref)) return 1;
    }

    err = TreoGetToneIDs(phonelibref, &ToneIDList, &ToneIDCount);

    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
void StopApplication(void)
{
    if (ToneIDList) MemPtrFree(ToneIDList);
    ClosePhoneLibrary(phonelibref);
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
void EventLoop(void)
{
    Err err;
    EventType e;
    Int32 eventTimeout;

    eventTimeout = 1*tps;

    do {

    EvtGetEvent(&e, eventTimeout);
    if (!SysHandleEvent(&e)) {
        if (!MenuHandleEvent(NULL, &e, &err)) {
            if (!ApplicationHandleEvent(&e)) {
                FrmDispatchEvent(&e);
            }
        }
    }

    } while (e.eType != appStopEvent);

}

Boolean is_os_32(void)
{
    UInt32 romVersion;
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    /* some NotifyMgr things need >= 3.2 */
    if (romVersion >= 0x03200000) return true;
    return false;
}

Boolean is_os_5x(void)
{
    UInt32 romVersion;
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    /* see if OS version is >= 5.0 */
    if (romVersion >= 0x05000000) return true;
    return false;
}

void register_notifyevent(UInt32 e)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (!is_os_32()) return;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
        sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
    SysNotifyRegister(cardNo, dbID, e, NULL, sysNotifyNormalPriority, NULL);
}

void unregister_notifyevent(UInt32 e)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (!is_os_32()) return;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
        sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
    SysNotifyUnregister(cardNo, dbID, e, sysNotifyNormalPriority);
}

/* From Handspring's PhoneLib Ref.pdf */

UInt32 WhichPhoneType(void)
{
    UInt32 phoneType;

    /* Check what device we are on: GSM or CDMA */
    (void)HsAttrGet(hsAttrPhoneType, 0, &phoneType);
    switch (phoneType) {
    case hsAttrPhoneTypeGSM:
    case hsAttrPhoneTypeCDMA:
        break;
    default:
        /* Pre Treo 300 phones should default to GSM */
        phoneType = hsAttrPhoneTypeGSM;
        break;
    }
    return phoneType;
}

Err LoadPhoneLibrary(UInt16 *plref)
{
    Err error=0;

    *plref = -1;

    /* Treo 600 changed to a new way.  WHY!  It broke all my stuff! */
    if (is_os_5x()) {
        error = HsGetPhoneLibrary(plref);
        return error;
    }

    switch (WhichPhoneType()) {
    case hsAttrPhoneTypeGSM:
        /* Get the Phone library */
        error = SysLibFind(phnLibName, plref);
        if (error) {
            /* It's not already here - have to load it ourselves */
            error = SysLibLoad(phnLibDbType, phnLibDbCreator, plref);
        }
        break;
    case hsAttrPhoneTypeCDMA:
        /* Get the Phone library */
        error = SysLibFind("PhoneLib.prc", plref);
        if (error) {
            /* It's not already here - have to load it ourselves */
            error = SysLibLoad(phnLibDbType, 'HsCL', plref);
        }
        break;
    }

    return error;
}

void ClosePhoneLibrary(UInt16 plref)
{
    if (plref != -1) {
        /* unregister for launch codes */
        //PhnLibRegister(plref, MainAppID, 0);
        PhnLibClose(plref);
    }
}

Err UseNetMasterLibrary(UInt16 *nmlref)
{
    Err error=0;

    /* Routine is pointless without this parameter */
    if (!nmlref) return memErrInvalidParam;
    /* Get the NetMaster Library */
    error = SysLibFind(netMasterLibName, nmlref);
    if (error) {
        /*
         * It's not already here - don't load it:
         * there's a good reason it's not loaded.
         */
    }

    return error;
}

void LoadPrefs(struct sprefs *prefs, UInt8 whichprefs)
{
    Int16 prefsVersion;
    UInt16 prefsSize;
    UInt32 ps;
    UInt16 i;
    Err err;

    prefs->phoneflags = 0;
    prefsSize = sizeof(prefs->phoneflags);
    prefsVersion = PrefGetAppPreferences(MainAppID, 0 /* Pref ID */,
        &prefs->phoneflags, &prefsSize, true);

    if (whichprefs < WHICHPREFS_DROPNET) return; /* so I don't load ALL prefs */

    prefs->phoneofftime = 0;
    prefsSize = sizeof(prefs->phoneofftime);
    prefsVersion = PrefGetAppPreferences(MainAppID, 2 /* Pref ID */,
        &prefs->phoneofftime, &prefsSize, true);

    prefs->dropnetgreentime = 0;
    prefsSize = sizeof(prefs->dropnetgreentime);
    prefsVersion = PrefGetAppPreferences(MainAppID, 5 /* Pref ID */,
        &prefs->dropnetgreentime, &prefsSize, true);

    prefs->dropnetgreytime = 0;
    prefsSize = sizeof(prefs->dropnetgreytime);
    prefsVersion = PrefGetAppPreferences(MainAppID, 6 /* Pref ID */,
        &prefs->dropnetgreytime, &prefsSize, true);

    prefs->dropnetcoverage = 0;
    prefsSize = sizeof(prefs->dropnetcoverage);
    prefsVersion = PrefGetAppPreferences(MainAppID, 7 /* Pref ID */,
        &prefs->dropnetcoverage, &prefsSize, true);

    if (whichprefs < WHICHPREFS_ALL) return;

    prefs->phoneturnon = 0xffff; /* my value for N/A */
    prefsSize = sizeof(prefs->phoneturnon);
    prefsVersion = PrefGetAppPreferences(MainAppID, 3 /* Pref ID */,
        &prefs->phoneturnon, &prefsSize, true);

    prefs->phoneturnoff = 0xffff; /* my value for N/A */
    prefsSize = sizeof(prefs->phoneturnoff);
    prefsVersion = PrefGetAppPreferences(MainAppID, 4 /* Pref ID */,
        &prefs->phoneturnoff, &prefsSize, true);

    prefs->phoneonofftype = 0; /* daily */
    prefsSize = sizeof(prefs->phoneonofftype);
    prefsVersion = PrefGetAppPreferences(MainAppID, 8 /* Pref ID */,
        &prefs->phoneonofftype, &prefsSize, true);

    for (i=0; i < 14; i++) prefs->phoneweekonoff[i] = 0xffff; /* N/A time */
    prefsSize = sizeof(prefs->phoneweekonoff);
    prefsVersion = PrefGetAppPreferences(MainAppID, 9 /* Pref ID */,
        &prefs->phoneweekonoff, &prefsSize, true);

    prefs->phoneontries = 1; /* one try */
    prefsSize = sizeof(prefs->phoneontries);
    prefsVersion = PrefGetAppPreferences(MainAppID, 10 /* Pref ID */,
        &prefs->phoneontries, &prefsSize, true);

    prefs->phoneonevery = 0; /* don't */
    prefsSize = sizeof(prefs->phoneonevery);
    prefsVersion = PrefGetAppPreferences(MainAppID, 11 /* Pref ID */,
        &prefs->phoneonevery, &prefsSize, true);

    for (i=0; i < SMSENTRIES; i++) {
        prefs->smsflags[i] = 0;
        prefs->smstrigtext[i][0] = (char)NULL;
        prefs->smslaunch[i] = 0;
        prefs->smstone[i] = 0;
        prefs->smsduration[i] = 3;
        prefs->smsvolume[i] = 1; /* soft */
    }
    prefsSize = sizeof(prefs->smsflags);
    prefsVersion = PrefGetAppPreferences(MainAppID, 13 /* Pref ID */,
        &prefs->smsflags, &prefsSize, true);

    prefsSize = sizeof(prefs->smstrigtext); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 14 /* Pref ID */,
        &prefs->smstrigtext, &prefsSize, true);

    prefsSize = sizeof(prefs->smslaunch); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 15 /* Pref ID */,
        &prefs->smslaunch, &prefsSize, true);

    prefsSize = sizeof(prefs->smstone); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 16 /* Pref ID */,
        &prefs->smstone, &prefsSize, true);

    prefsSize = sizeof(prefs->smsduration); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 17 /* Pref ID */,
        &prefs->smsduration, &prefsSize, true);

    prefsSize = sizeof(prefs->smsvolume); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 18 /* Pref ID */,
        &prefs->smsvolume, &prefsSize, true);

    for (i=0; i < FUNENTRIES; i++) {
        prefs->funtone[i] = 0;
        prefs->funduration[i] = 3;
        prefs->funvolume[i] = 1; /* soft */
    }

    prefsSize = sizeof(prefs->funtone); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 19 /* Pref ID */,
        &prefs->funtone, &prefsSize, true);

    prefsSize = sizeof(prefs->funduration); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 20 /* Pref ID */,
        &prefs->funduration, &prefsSize, true);

    prefsSize = sizeof(prefs->funvolume); /* I defaulted it above */
    prefsVersion = PrefGetAppPreferences(MainAppID, 21 /* Pref ID */,
        &prefs->funvolume, &prefsSize, true);

    prefs->funhandsfreenumber[0] = (char)NULL;
    prefsSize = sizeof(prefs->funhandsfreenumber);
    prefsVersion = PrefGetAppPreferences(MainAppID, 22 /* Pref ID */,
        &prefs->funhandsfreenumber, &prefsSize, true);

    /* retrieve current value from the Handspring prefs */
    prefs->lidofftime = 30;
    ps = sizeof(prefs->lidofftime);
    err = HsPrefGet(hsPrefLidClosedAutoOffSeconds, &prefs->lidofftime,
        &ps);
}

Boolean LoadTempPrefs(struct sprefs **prefs, UInt8 whichprefs)
{
    if ((*prefs = MemPtrNew(sizeof(struct sprefs)))) {
        LoadPrefs(*prefs, whichprefs);
    }
    return (*prefs != NULL);
}

void UnloadTempPrefs(struct sprefs *prefs)
{
    MemPtrFree(prefs);
}

void SavePrefs(struct sprefs *prefs)
{
    PrefSetAppPreferences(MainAppID, 0 /* Pref ID */, 1 /* Version */,
        &prefs->phoneflags, sizeof(prefs->phoneflags), true);

    PrefSetAppPreferences(MainAppID, 2 /* Pref ID */, 1 /* Version */,
        &prefs->phoneofftime, sizeof(prefs->phoneofftime), true);

    PrefSetAppPreferences(MainAppID, 3 /* Pref ID */, 1 /* Version */,
        &prefs->phoneturnon, sizeof(prefs->phoneturnon), true);

    PrefSetAppPreferences(MainAppID, 4 /* Pref ID */, 1 /* Version */,
        &prefs->phoneturnoff, sizeof(prefs->phoneturnoff), true);

    PrefSetAppPreferences(MainAppID, 5 /* Pref ID */, 1 /* Version */,
        &prefs->dropnetgreentime, sizeof(prefs->dropnetgreentime), true);

    PrefSetAppPreferences(MainAppID, 6 /* Pref ID */, 1 /* Version */,
        &prefs->dropnetgreytime, sizeof(prefs->dropnetgreytime), true);

    PrefSetAppPreferences(MainAppID, 7 /* Pref ID */, 1 /* Version */,
        &prefs->dropnetcoverage, sizeof(prefs->dropnetcoverage), true);

    PrefSetAppPreferences(MainAppID, 8 /* Pref ID */, 1 /* Version */,
        &prefs->phoneonofftype, sizeof(prefs->phoneonofftype), true);

    PrefSetAppPreferences(MainAppID, 9 /* Pref ID */, 1 /* Version */,
        &prefs->phoneweekonoff, sizeof(prefs->phoneweekonoff), true);

    PrefSetAppPreferences(MainAppID, 10 /* Pref ID */, 1 /* Version */,
        &prefs->phoneontries, sizeof(prefs->phoneontries), true);

    PrefSetAppPreferences(MainAppID, 11 /* Pref ID */, 1 /* Version */,
        &prefs->phoneonevery, sizeof(prefs->phoneonevery), true);

    PrefSetAppPreferences(MainAppID, 13 /* Pref ID */, 1 /* Version */,
        &prefs->smsflags, sizeof(prefs->smsflags), true);

    PrefSetAppPreferences(MainAppID, 14 /* Pref ID */, 1 /* Version */,
        &prefs->smstrigtext, sizeof(prefs->smstrigtext), true);

    PrefSetAppPreferences(MainAppID, 15 /* Pref ID */, 1 /* Version */,
        &prefs->smslaunch, sizeof(prefs->smslaunch), true);

    PrefSetAppPreferences(MainAppID, 16 /* Pref ID */, 1 /* Version */,
        &prefs->smstone, sizeof(prefs->smstone), true);

    PrefSetAppPreferences(MainAppID, 17 /* Pref ID */, 1 /* Version */,
        &prefs->smsduration, sizeof(prefs->smsduration), true);

    PrefSetAppPreferences(MainAppID, 18 /* Pref ID */, 1 /* Version */,
        &prefs->smsvolume, sizeof(prefs->smsvolume), true);

    PrefSetAppPreferences(MainAppID, 19 /* Pref ID */, 1 /* Version */,
        &prefs->funtone, sizeof(prefs->funtone), true);

    PrefSetAppPreferences(MainAppID, 20 /* Pref ID */, 1 /* Version */,
        &prefs->funduration, sizeof(prefs->funduration), true);

    PrefSetAppPreferences(MainAppID, 21 /* Pref ID */, 1 /* Version */,
        &prefs->funvolume, sizeof(prefs->funvolume), true);

    PrefSetAppPreferences(MainAppID, 22 /* Pref ID */, 1 /* Version */,
        &prefs->funhandsfreenumber, sizeof(prefs->funhandsfreenumber), true);

    /* don't set the lid closed auto off for Treo 600 */
    if (!is_os_5x()) {
        /* save current value to the Handspring prefs */
        HsPrefSet(hsPrefLidClosedAutoOffSeconds, &prefs->lidofftime,
            sizeof(prefs->lidofftime));
    }
}

void SetTimeOfNextAlarm(UInt32 alarmTime, UInt32 ref, AlmAlarmProcPtr procP)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (procP) {
        AlmSetProcAlarm(procP, ref, alarmTime);
    } else {
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
        AlmSetAlarm(cardNo, dbID, ref, alarmTime, 0);
    }
}

UInt32 GetTimeOfNextAlarm(UInt32 *refP, AlmAlarmProcPtr procP)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (procP) {
        return AlmGetProcAlarm(procP, refP);
    } else {
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return 0;
        return AlmGetAlarm(cardNo, dbID, refP);
    }
}

void sync_autologin_setting(struct sprefs *tprefs, Boolean forceoff)
{
    UInt16 tnmlibref;
    NetMasterAutoLoginType nmal;

    /*
     * There is a system preference for auto-login when the phone is turned
     * on.  On the CDMA Treo this works, but on the GSM/GPRS it does sortof
     * a half-baked login and doesn't do the network connection.  Why?  I dunno.
     *
     * Also the system pref gets cleared on some systems after a reset.  So we
     * have a local pref for it, and what we do is sync it with the system pref
     * after every reset or local pref change.  Well, basically we do a one-way
     * sync.  Our setting will write over the system pref.
     */
    if (!UseNetMasterLibrary(&tnmlibref) &&
        !NetMasterAutoLoginSettingGet(tnmlibref, &nmal)) {
        if ((tprefs->phoneflags&PHONEFLAGS_AUTOLOGIN) &&
            !(nmal&netMasterAutoLoginWhenWirelessOn)) {
            /*
             * Some users have reported that the system preference for
             * auto-login doesn't stick across resets.  So if our pref
             * is set, we'll set the system's as well.
             */
            /* set it */
            NetMasterAutoLoginSettingSet(tnmlibref, 0,
                netMasterAutoLoginWhenWirelessOn);
        } else if (forceoff &&
            !(tprefs->phoneflags&PHONEFLAGS_AUTOLOGIN) &&
            (nmal&netMasterAutoLoginWhenWirelessOn)) {
            /* clear it */
            NetMasterAutoLoginSettingSet(tnmlibref,
                netMasterAutoLoginWhenWirelessOn, 0);
        }
    }
}

void phone_events_register(UInt16 plref, struct sprefs *tprefs)
{
    UInt16 tui16;

    /* register for phone event launch codes */
    tui16 = phnServiceSMS|phnServiceData;
    /*
     * Normally I don't need phnServiceVoice events but for the
     * trick I use to dismiss an error dialog, I need it.  That's
     * because I have found no other way to get a clean event
     * when the error dialog occurs.
     *
     * The problem is that there are some (all) 270 versions which
     * have a problem answering/hangingup with the handfree button.
     * We tracked it down, and it is basically if you register for
     * phnServiceVoice events.  Even if you just "return 0" in the
     * PilotMain() it still messes up.
     *
     * So that's why I only register for these events if you pick
     * the auto-ok option.  That way when someone complains I can
     * just tell them to turn it off and reboot.
     *
     * Same is true for handsfree auto-dial.  It requires
     * phnServiceVoice events phone to detect the handsfree button.
     */
    if ((tprefs->phoneflags&PHONEFLAGS_AUTOOKCONNECT) ||
        (tprefs->phoneflags&PHONEFLAGS_ENABLEHANDSFREEDIAL))
        tui16 |= phnServiceVoice;
    PhnLibRegister(plref, MainAppID, tui16);
}

void phone_events_unregister(UInt16 plref)
{
    UInt16 tui16;

    tui16 = 0;
    PhnLibRegister(plref, MainAppID, tui16);
}

void AlarmProcMulti(UInt16 almProcCmd, SysAlarmTriggeredParamType *paramP)
{
    if (almProcCmd != almProcCmdTriggered) return;
    GenericAlarm(paramP->ref, paramP->alarmSeconds);
}

void DisplayAlarm(SysDisplayAlarmParamType *paramP)
{
    GenericAlarm(paramP->ref, paramP->alarmSeconds);
}

void GenericAlarm(UInt32 ref, UInt32 alarmSeconds)
{
    UInt16 tplibref, tanref;
    UInt32 tui32;
    Err err;
    UInt16 i;
    char *p;
    struct sprefs *tprefs;

    /* set our ftr to zero so it doesn't cause another alarm */
    FtrSet(MainAppID, ref, 0);

    if (!LoadTempPrefs(&tprefs, WHICHPREFS_ALL)) { /* failed to load prefs? */
        set_next_alarm_ref();
        return;
    }

    switch (ref) {

    case FTR_DROPNETGREENTIME:
    case FTR_DROPNETGREYTIME:
    case FTR_DROPNETCOVERAGE:
        /* drop 3G network connection (i.e. arrows) */
        HsNetworkDropConnection();
        break;

    case FTR_AUTOLOGIN:
        /*
         * GSM phones don't auto-login to the network.  Well, there
         * is a setting in Preferences->Network->Options for it, but
         * it doesn't actually do a "Service Connection Progress".
         *
         * So I am looking at the pref and I am going to do a network
         * connection when we come out of registration search.  I'll
         * only do this once per power-on -- that's what my feature
         * is for.
         *
         * If we arrive in this timer, it's because a) we are GSM, and b)
         * because both the system's auto-login is set, and our's is.
         */
        if (!SysLibFind("Net.lib", &tanref)) {
            /* open/close netlib to make it login */
            NetLibOpen(tanref, &err);
            NetLibClose(tanref, false); /* no hangup */
        }
        break;

    case FTR_AUTOOKCONNECT:
        /*
         * This is a timer we set when we recognize a Connection Progress
         * dialog.  We use Handspring's NetMaster API to determine the
         * state of the login attempt, and auto-ok an error dialog if
         * present.
         */
        dismiss_progress_dialog_generic();
        break;

    case FTR_STOPTONE:
        /*
         * If you use loud volume on a Treo 300 ringtone, the amp will
         * get stuck somehow.  For example if you go to make a call, it
         * will seem like the speakerphone is on.  So I set this timer
         * for when the ringtone duration is over, and I correct the
         * volume.
         */
        if ((WhichPhoneType() == hsAttrPhoneTypeCDMA) &&
            !LoadPhoneLibrary(&tplibref)) {
            PhnLibStopTone(tplibref);
            ClosePhoneLibrary(tplibref);
        }
        break;

    case FTR_COVERAGELOSS: /* turn off phone due to coverage loss timer */
        EvtResetAutoOffTimer();
        if (!LoadPhoneLibrary(&tplibref)) {
            PhnLibSetModulePower(tplibref, false);
            ClosePhoneLibrary(tplibref);
            /* schedule a "phone on" attempt */
            if (!FtrGet(MainAppID, FTR_PHONEONTRIES, &tui32) &&
                (tui32 < tprefs->phoneontries)) {
                tui32 = tprefs->phoneonevery;
                if (tui32 > 0) {
                    tui32 *= 60; /* turn into seconds */
                    tui32 += TimGetSeconds();
                    FtrSet(MainAppID, FTR_PHONEONEVERY, tui32);
                }
            }
        }
        EvtSetAutoOffTimer(SetExactly, 10); /* power off in 10 seconds */
        break;

    case FTR_RESET: /* delayed reset */
        /* overwrite system pref, but don't force it off */
        sync_autologin_setting(tprefs, false);
        /* no break wanted here */
    case FTR_PHONEONEVERY:
        EvtResetAutoOffTimer();
        /* increment "phone on" attempts */
        if (!FtrGet(MainAppID, FTR_PHONEONTRIES, &tui32) &&
            (tui32 < tprefs->phoneontries)) {
            tui32++;
            FtrSet(MainAppID, FTR_PHONEONTRIES, tui32);
        }
        if (!LoadPhoneLibrary(&tplibref)) {
            if (ref == FTR_RESET) {
                phone_events_register(tplibref, tprefs);
            }
            /*
             * should_phone_be_on() gives a liberal idea of when it is
             * appropriate for the phone to be on based on the user's
             * on/off schedule.  Note that "enable phone on reset" is
             * independent of the coverage loss retry attempts.  It was
             * broken in 0.11.
             */
            if (((tprefs->phoneflags&PHONEFLAGS_ENABLEPHONEONRESET) ||
                (ref == FTR_PHONEONEVERY))) {
                /* FTR_RESET checks if phone should be on */
                if ((ref == FTR_RESET) &&
                    !should_phone_be_on(tprefs, alarmSeconds)) {}
                else {
                    if (!PhnLibModulePowered(tplibref)) {
                        PhnLibSetModulePower(tplibref, true);
                    }
                }
            }
            ClosePhoneLibrary(tplibref);
        }
        EvtSetAutoOffTimer(SetExactly, 10); /* power off in 10 seconds */
        break;

    case FTR_PHONETURNON: /* phoneturnon */
        EvtResetAutoOffTimer();
        if (!LoadPhoneLibrary(&tplibref)) {
            if (!PhnLibModulePowered(tplibref))
                PhnLibSetModulePower(tplibref, true);
            ClosePhoneLibrary(tplibref);
        }
        set_phoneonoff_times(tprefs, alarmSeconds); /* for tomorrow */
        EvtSetAutoOffTimer(SetExactly, 10); /* power off in 10 seconds */
        break;

    case FTR_PHONETURNOFF: /* phoneturnoff */
        EvtResetAutoOffTimer();
        if (!LoadPhoneLibrary(&tplibref)) {
            /*
             * First off, let's make sure the user isn't in the middle of a
             * phone call.  If they are, then just forget about turning off
             * the phone.
             */
            if (PhnLibModulePowered(tplibref) &&
                !PhnLibGetPhoneCallStatus(tplibref, &tui32) &&
                !(tui32&(phnVoiceCall1Active|phnVoiceCall2Active))) {
                PhnLibSetModulePower(tplibref, false);
            }
            ClosePhoneLibrary(tplibref);
        }
        /* also cancel "phone on" attempts */
        FtrSet(MainAppID, FTR_PHONEONEVERY, 0); /* interval timer */
        set_phoneonoff_times(tprefs, alarmSeconds); /* for tomorrow */
        EvtSetAutoOffTimer(SetExactly, 10); /* power off in 10 seconds */
        break;

    case FTR_DELAYLAUNCH:
        for (i=0; i < SMSENTRIES; i++) {
            /* see if there is feature memory alloc'd in any SMS slot */
            if (!FtrGet(MainAppID, FTR_SMS1+i, (UInt32 *)&p)) {
                /*
                 * Okay, I found one so we launch just this one.  If there
                 * is more than one, it will be picked up the next time we
                 * sleep.
                 */
                do_sms_launch(tprefs, p, i);
                /* erase the feature memory because we use it as a flag */
                FtrPtrFree(MainAppID, FTR_SMS1+i);
                break;
            }
        }
        break;

    case FTR_HANDSFREEDIAL:
        /*
         * The user has pressed the handsfree button.  I use a display alarm
         * to initiate it because I can't do it in the PhoneLib callback.
         */
        dial_phone(tprefs->funhandsfreenumber);
        break;

    }

    set_next_alarm_ref();
    UnloadTempPrefs(tprefs);
}

void AlarmTriggered(SysAlarmTriggeredParamType *paramP)
{
    switch (paramP->ref) {

    case FTR_PHONEONEVERY:
    case FTR_PHONETURNON:
    case FTR_PHONETURNOFF:
    case FTR_HANDSFREEDIAL:
        /*
         * Dismiss a Progress Connection Dialog if present.  Otherwise it
         * could prevent other stuff from happening.
         */
        dismiss_progress_dialog();
        break;

    }

    /* if purged, set our ftr to zero so it doesn't cause another alarm */
    if (paramP->purgeAlarm) FtrSet(MainAppID, paramP->ref, 0);
}

void dismiss_progress_dialog(void)
{
    UInt16 tnmlibref;
    Boolean isLogonActive, isPrgDialogUp, waitingInErrorState;
    /*
     * Use Handspring's NetMaster API to determine if there is an error
     * dialog, and auto-ok it.  Otherwise the DisplayAlarm won't fire.
     * This is important for when I want to turn the phone on/off.
     */
    if (!UseNetMasterLibrary(&tnmlibref)) {
        if (NetMasterNetLibIsLoggingIn(tnmlibref, &isLogonActive,
            &isPrgDialogUp, &waitingInErrorState)) {
            /* it is logging in */
            if (waitingInErrorState) {
                /* tell it to get rid of the dialog */
                NetMasterNetInterfacesShutDown(tnmlibref, 0,
                    netMasterNetIFShutDownOptDismissPrgDialog);
            }
        }
    }
}

void dismiss_progress_dialog_generic(void)
{
    FormPtr frm;
    ControlPtr ptr;
    UInt16 numobj;
    Int16 i1;
    char *lab;

    /*
     * I use this to click-through those annoying network
     * dialogs.
     */
    frm = FrmGetActiveForm();
    if (!frm) return;
    if (!FrmVisible(frm)) return;
    lab = (char *)FrmGetTitle(frm);
    if (lab &&
        (StrCaselessCompare("Service Connection Progress", lab) == 0)) {
        numobj = FrmGetNumberOfObjects(frm);
        for (i1=0; i1 < numobj; i1++) {
            ptr = FrmGetObjectPtr(frm, i1);
            if ((ptr->attr.enabled) && (ptr->style == buttonCtl)) {
                lab = (char *)CtlGetLabel(ptr);
                if (lab && (StrCaselessCompare("OK", lab) == 0)) {
                    CtlHitControl(ptr);
                    EvtWakeup(); /* make sure the event is processed */
                    break;
                }
            }
        }
    }
}

/*
 * Since only one alarm at a time is allowed.  We find the closest one and
 * set it.
 */
void set_next_alarm_ref(void)
{
    UInt32 lowtime, ftrtime, ref=0xffff, ref2;
    Boolean procalarm=true;
    AlmAlarmProcPtr pa=NULL;
    UInt32 curtime;
    UInt32 wpt;
    AlmAlarmProcPtr oldfunc;

    /*
     * Most things work better in a procalarm.  It doesn't turn the screen on,
     * and it isn't stopped by an alert dialog.  Some things don't work so well
     * in a procalarm though, and it's different depending on the device.
     *
     * For example, GSM devices have intermittent problems with turning the
     * phone on/off in a procalarm.  And handsfree dialing doesn't work from
     * a procalarm either.
     */
    wpt = WhichPhoneType();

    curtime = 1; /* we want to catch all we've set > 0 */

    if (FtrGet(MainAppID, FTR_COVERAGELOSS, &ftrtime)) ftrtime = 0;
    lowtime = 0xffffffff;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_COVERAGELOSS;
        procalarm = (wpt == hsAttrPhoneTypeCDMA) ? true : false;
        if (is_os_5x()) procalarm = true; /* procalarms OK on Treo 600 */
    }
    if (FtrGet(MainAppID, FTR_RESET, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_RESET;
        procalarm = (wpt == hsAttrPhoneTypeCDMA) ? true : false;
        if (is_os_5x()) procalarm = true; /* procalarms OK on Treo 600 */
    }
    if (FtrGet(MainAppID, FTR_PHONETURNON, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_PHONETURNON;
        procalarm = (wpt == hsAttrPhoneTypeCDMA) ? true : false;
        if (is_os_5x()) procalarm = true; /* procalarms OK on Treo 600 */
    }
    if (FtrGet(MainAppID, FTR_PHONETURNOFF, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_PHONETURNOFF;
        procalarm = (wpt == hsAttrPhoneTypeCDMA) ? true : false;
        if (is_os_5x()) procalarm = true; /* procalarms OK on Treo 600 */
    }
    if (FtrGet(MainAppID, FTR_PHONEONEVERY, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_PHONEONEVERY;
        procalarm = (wpt == hsAttrPhoneTypeCDMA) ? true : false;
        if (is_os_5x()) procalarm = true; /* procalarms OK on Treo 600 */
    }
    if (FtrGet(MainAppID, FTR_DROPNETGREENTIME, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_DROPNETGREENTIME;
        procalarm = true;
    }
    if (FtrGet(MainAppID, FTR_DROPNETGREYTIME, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_DROPNETGREYTIME;
        procalarm = true;
    }
    if (FtrGet(MainAppID, FTR_DROPNETCOVERAGE, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_DROPNETCOVERAGE;
        procalarm = true;
    }
    if (FtrGet(MainAppID, FTR_AUTOOKCONNECT, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_AUTOOKCONNECT;
        procalarm = true;
    }
    if (FtrGet(MainAppID, FTR_AUTOLOGIN, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_AUTOLOGIN;
        procalarm = true;
    }
    if (FtrGet(MainAppID, FTR_STOPTONE, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_STOPTONE;
        procalarm = true;
    }
    if (FtrGet(MainAppID, FTR_DELAYLAUNCH, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_DELAYLAUNCH;
        procalarm = false;
    }
    if (FtrGet(MainAppID, FTR_HANDSFREEDIAL, &ftrtime)) ftrtime = 0;
    if ((ftrtime >= curtime) && /* should be in future */
        (ftrtime < lowtime)) {
        lowtime = ftrtime;
        ref = FTR_HANDSFREEDIAL;
        procalarm = false;
    }

    /*
     * I'd prefer to do everything via procedure alarm, but some things
     * don't work correctly.  So I use a regular alarm for them.
     */
    if (procalarm) pa = AlarmProcMulti;

    /* clear procalarm if any */
    if (FtrGet(MainAppID, FTR_PROCALARM_FUNC, (void *)&oldfunc) == 0) {
        if (GetTimeOfNextAlarm(&ref2, oldfunc))
            SetTimeOfNextAlarm(0, ref2, oldfunc);
    }
    /* clear regular alarm if any */
    if (GetTimeOfNextAlarm(&ref2, 0))
        SetTimeOfNextAlarm(0, ref2, NULL);

    if (ref != 0xffff) {
        /* store the function ptr for canceling - especially at hotsync */
        FtrSet(MainAppID, FTR_PROCALARM_FUNC, (UInt32)pa);
        SetTimeOfNextAlarm(lowtime, ref, pa);
    }
}

UInt32 midnight_seconds(UInt32 tgs)
{
    DateTimeType dt;
    TimSecondsToDateTime(tgs, &dt);
    /* calculate how many seconds since midnight */
    return (((UInt32)dt.hour*3600)+((UInt32)dt.minute*60)+(UInt32)dt.second);
}

UInt32 monday_seconds(UInt32 tgs)
{
    DateTimeType dt;
    UInt32 ts;
    Int16 wd;
    TimSecondsToDateTime(tgs, &dt);
    /* convert from Sunday based to Monday based */
    wd = dt.weekDay-1;
    if (wd < 0) wd = 6;
    /* calculate how many seconds since Monday */
    ts = ((UInt32)dt.hour*3600)+((UInt32)dt.minute*60)+(UInt32)dt.second+
        ((UInt32)wd*24*3600);
    return ts;
}

void set_phoneonoff_times(struct sprefs *tprefs, UInt32 tgs)
{
    UInt32 daysec, weeksec;
    UInt32 onmin=0xffff, offmin=0xffff, onsec=0, offsec=0;
    UInt32 ton, toff, tons, toffs;
    UInt16 i;

    /* I may have called this with a specific time, otherwise use now */
    if (tgs == 0) tgs = TimGetSeconds();

    daysec = midnight_seconds(tgs);
    weeksec = monday_seconds(tgs);

    switch (tprefs->phoneonofftype) {
    case 0: /* daily */
        onmin = tprefs->phoneturnon;
        offmin = tprefs->phoneturnoff;
        onsec = onmin*60;
        offsec = offmin*60;
        /* adjust for next day if necessary */
        onsec = (daysec < onsec) ? onsec-daysec : (onsec+86400)-daysec;
        offsec = (daysec < offsec) ? offsec-daysec : (offsec+86400)-daysec;
        break;
    case 1: /* weekly */
        /* find the next on/off time during the week */
        for (i=0; i < 7; i++) {
            ton = tprefs->phoneweekonoff[i];
            if (ton != 0xffff) {
                /* bias by day of week */
                ton = ton+(i*(24*60));
                tons = ton*60;
                /* assign the first valid on time in the future */
                if ((onmin == 0xffff) && (weeksec < tons)) {
                    onmin = ton;
                    onsec = tons-weeksec;
                }
            }
            toff = tprefs->phoneweekonoff[i+7];
            if (toff != 0xffff) {
                /* bias by day of week */
                toff = toff+(i*(24*60));
                toffs = toff*60;
                /* assign the first valid off time in the future */
                if ((offmin == 0xffff) && (weeksec < toffs)) {
                    offmin = toff;
                    offsec = toffs-weeksec;
                }
            }
        }
        /*
         * A trick.  We need to find a future on and off time.  But we
         * also want to rollover to the following week; i.e. off time is
         * on Friday, and on time is Monday.  So it needs to span two weeks
         * in order to find the appropriate times.
         */
        if (onmin == 0xffff) {
            /* find first valid on time in the next week */
            for (i=0; i < 7; i++) {
                ton = tprefs->phoneweekonoff[i];
                if (ton != 0xffff) {
                    /* bias by day of week */
                    ton = ton+(i*(24*60));
                    tons = ton*60;
                    onmin = ton; /* just to set it non 0xffff */
                    onsec = tons+(604800/*secs per week*/-weeksec);
                    break;
                }
            }
        }
        if (offmin == 0xffff) {
            /* find first valid off time in the next week */
            for (i=0; i < 7; i++) {
                toff = tprefs->phoneweekonoff[i+7];
                if (toff != 0xffff) {
                    /* bias by day of week */
                    toff = toff+(i*(24*60));
                    toffs = toff*60;
                    offmin = toff; /* just to set it non 0xffff */
                    offsec = toffs+(604800/*secs per week*/-weeksec);
                    break;
                }
            }
        }
        break;
    }

    /* this value will be in minutes (0-20160), or 0xffff if N/A */
    if (onmin != 0xffff) {
        FtrSet(MainAppID, FTR_PHONETURNON, tgs+onsec);
    } else {
        /* cancel the alarm */
        FtrSet(MainAppID, FTR_PHONETURNON, 0);
    }
    /* this value will be in minutes (0-20160), or 0xffff if N/A */
    if (offmin != 0xffff) {
        FtrSet(MainAppID, FTR_PHONETURNOFF, tgs+offsec);
    } else {
        /* cancel the alarm */
        FtrSet(MainAppID, FTR_PHONETURNOFF, 0);
    }

    set_next_alarm_ref();
}

/*
 * This application is an exploration on how to work-around various radio
 * bugs on the Treo.  One of the common problems is when the network arrows
 * get stuck on green, and cause a super quick battery drain.  So we have
 * a timer which determines the longest continuous green arrow, after which
 * we drop the whole network connection.  This should be set for a length
 * of time which is not likely, like 5 minutes or so.
 *
 * The grey arrow timer is just a feature which should have been in the prefs
 * in the first place.  A way to disconnect after a certain idle time.
 */
void set_dropnet_times(struct sprefs *tprefs, UInt32 tgs)
{
    UInt32 secs, tui32;
    UInt16 tplibref;

    /* I may have called this with a specific time, otherwise use now */
    if (tgs == 0) tgs = TimGetSeconds();

    /*
     * We set or reset the appropriate timer depending on the
     * current color of the arrows.
     */
    if (!LoadPhoneLibrary(&tplibref)) {
        if (!PhnLibGetPhoneCallStatus(tplibref, &tui32)) {
            /* are 3G arrows green? */
            if (tui32&phnGPRSCallActive) {
                /* yup, set our green time reaper */
                secs = (tprefs->dropnetgreentime > 0) ?
                    (tgs+tprefs->dropnetgreentime) : 0;
                FtrSet(MainAppID, FTR_DROPNETGREENTIME, secs);
                FtrSet(MainAppID, FTR_DROPNETGREYTIME, 0); /* clear grey */
            } else if (tui32&0x0010 /* 3G in idle (grey arrows) */) {
                secs = (tprefs->dropnetgreytime > 0) ?
                    (tgs+tprefs->dropnetgreytime) : 0;
                FtrSet(MainAppID, FTR_DROPNETGREYTIME, secs);
                FtrSet(MainAppID, FTR_DROPNETGREENTIME, 0); /* clear green */
            }
        }
        ClosePhoneLibrary(tplibref);
    }
    set_next_alarm_ref();
}

Boolean should_phone_be_on(struct sprefs *tprefs, UInt32 tgs)
{
    UInt32 onmin, offmin;
    UInt32 daymin, weekmin;
    UInt16 i;
    Boolean ison=true;

    /* I may have called this with a specific time, otherwise use now */
    if (tgs == 0) tgs = TimGetSeconds();

    switch (tprefs->phoneonofftype) {
    case 0: /* daily */
        daymin = midnight_seconds(tgs)/60;
        /*
         * Thinking about this is a little weird.  If you set on time to N/A,
         * you could assume that we don't want it on after a reset.  But then
         * again why would you have enabled the reset option?  So, since my own
         * preference is to have on time as N/A, off time to midnight, and turn
         * the reset option on.  The way I expect it to work, is that I manually
         * turn the phone on, except that during a reset it will turn itself on.
         */
        onmin = tprefs->phoneturnon;
        offmin = tprefs->phoneturnoff;
        if (onmin == 0xffff) break; /* no on time, then should be on */
        if (offmin == 0xffff) break; /* no off time, then should be on */
        ison = false;
        /* if both times are set, falls within on/off range? */
        if (onmin < offmin) { /* normal AM-->PM range */
            if (daymin >= onmin) ison = true;
            if (daymin >= offmin) ison = false;
        } else { /* weird midnight wrapping range; i.e. on=8AM off=2AM */
            if (daymin < offmin) ison = true;
            if (daymin >= onmin) ison = true;
        }
        break;
    case 1: /* weekly */
        weekmin = monday_seconds(tgs)/60;

        /*
         * Now we're checking to see if there is an open ended off time.
         * Keep in mind that this function is only used to determine whether
         * the user would like the phone on after a reset.  So it implies
         * some "on" times based on the on/off times which have been chosen.
         *
         * For example, suppose the user has "Phone on after reset" checked,
         * but has no on/off times.  Well then of course he wants the phone to
         * be turned on after a reset regardless of time.  Other cases might
         * be where only off times are specified, but no on times.
         *
         * In fact, the only time when the "Phone on after reset" is not
         * acted upon, is when there is a specific block of on/off which
         * isn't left open (i.e. off: N/A).
         *
         * Consider this; the whole week is N/A except for an "on" time
         * for Thursday.  My take on that is that on Thursday the phone
         * will come on via schedule, and any resets which occur should
         * always turn the phone on if "Phone on after reset" is enabled.
         * That means I need to recognize on Monday, for example, that
         * there is an open-ended "on" time somewhere in the week.
         *
         * So I just crawl backwards through the week schedule and see if
         * there is an "on" without a terminating "off".
         */
        for (i=6; i >= 0; i--) {
            onmin = tprefs->phoneweekonoff[i];
            offmin = tprefs->phoneweekonoff[i+7];
            if ((offmin != 0xffff) &&
                (onmin < offmin)) { /* week is not open-ended */
                /*
                 * Also note that onmin<offmin gives me what I want
                 * in regard to when no "on" time is specified, to
                 * make it always "Phone on after reset" if enabled.
                 * This is because onmin=0xffff and therefore it
                 * won't allow it to think it's not open-ended.
                 */
                ison = false; /* default to phone off after reset */
                break;
            }
            if (onmin != 0xffff) break; /* week is open-ended */
        }

        /* Run through the week of on/off times */
        for (i=0; i < 7; i++) {
            onmin = tprefs->phoneweekonoff[i];
            offmin = tprefs->phoneweekonoff[i+7];
            /* bias them to the proper day of week */
            if (onmin != 0xffff) onmin += (i*(24*60));
            if (offmin != 0xffff) offmin += (i*(24*60));
            if (onmin < offmin) {
                /* range is --->on--->off */
                if (weekmin >= onmin) ison = true;
                if (weekmin >= offmin) ison = false;
            } else {
                /* range is off--->on---> */
                if (weekmin >= offmin) ison = false;
                if (weekmin >= onmin) ison = true;
            }
        }
    }

    return ison;
}

/*
 * This knowledge comes mostly from Alan Penner's SMSChat and SimpleSMS.c
 */
Boolean HandleSMS(UInt32 messageID)
{
    PhnAddressList addList;
    PhnAddressHandle addressH;
    MemHandle messageH;
    char *addressP;
    char *messageP;
    Boolean ret=false;
    Err err;
    UInt16 tplibref;
    struct sprefs *tprefs;
    Int16 idx=0;
    DmOpenRef smsRefNum=0; /* CDMA workaround??? */

//    MemPtr startPP, endPP;
//    SysGetStackInfo(&startPP, &endPP);
//    ErrFatalDisplayIf((endPP-startPP) < 4000, "HandleSMS stack < 4000");

    if (!LoadTempPrefs(&tprefs, WHICHPREFS_ALL)) return false;

    if (!LoadPhoneLibrary(&tplibref)) {
        smsRefNum = PhnLibGetDBRef(tplibref); /* CDMA workaround??? */
        do {
        /* get the address list */
        err = PhnLibGetAddresses(tplibref, messageID, &addList);
        if (err) break;
        /* now get the first entry from the list */
        err = PhnLibGetNth(tplibref, addList, 1, &addressH);
        MemHandleFree(addList); /* dispose of list */
        if (err) break;
        /* now get the actual address */
        addressP = PhnLibGetField(tplibref, addressH, phnAddrFldPhone);
        MemHandleFree(addressH); /* dispose of nth */
        if (addressP) MemPtrFree(addressP);
        /* and get the message text */
        err = PhnLibGetText(tplibref, messageID, &messageH);
        if (err) break;
        messageP = MemHandleLock(messageH);
        /*
         * There may be some messages which we should ignore.  Add tests for
         * them here.  VeriChat does its own sounds etc, so we don't want to
         * interfere.
         */
        if (StrStr(messageP, ".verichat.com:")) {
            /* VeriChat message -- ignore */
        } else if (StrStr(messageP, "shortmail@oneboxmail.bulk.sprint:")) {
            /* Shortmail trigger -- ignore */
        } else if ((idx = MatchSMStext(tplibref, tprefs, messageP)) != -1) {
            ret = true;
        }
        if (messageH) {
            MemHandleUnlock(messageH);
            MemHandleFree(messageH);
        }
        /* delete sms message if we matched */
        if (ret && (tprefs->smsflags[idx]&SMSFLAGS_DELETE)) {
            PhnLibDeleteMessage(tplibref, messageID, 0);
        } else {
            /* when not deleted, I want it to be processed by others */
            ret = false;
        }
        } while (false);
        PhnLibReleaseDBRef(tplibref, smsRefNum); /* CDMA workaround??? */
        ClosePhoneLibrary(tplibref);
    }

    UnloadTempPrefs(tprefs);
    return ret;
}

/*
 * This is HORRIBLE!  Part of the PhoneLib is redefined for the 300, with
 * overlapping API and different traps.  I don't see how to get the ringtone
 * API to work smoothly between the GSM and CDMA.  So I'm making functions
 * which call the appropriate trap.
 */

Err TreoGetToneIDs(UInt16 refNum, void **list, UInt16 *listLength)
{
    if (!PhnLibModulePowered(refNum)) return 1; /* sanity */
    switch (WhichPhoneType()) {
    case hsAttrPhoneTypeGSM:
        return PhnLibGetToneIDs(refNum, (UInt32 **)list, listLength);
    case hsAttrPhoneTypeCDMA:
        return Treo300PhnLibGetToneIDs(refNum, 7 /* phnRingerAudioTypeAll */,
            list, listLength);
    }
    return 1;
}

Err TreoGetToneName(UInt16 refNum, UInt32 tone, char *name,
    UInt16 maxLength)
{
    UInt16 i;

    if (!PhnLibModulePowered(refNum)) return 1; /* sanity */
    switch (WhichPhoneType()) {
    case hsAttrPhoneTypeGSM:
        /* for GSM versions, find the tone's index */
        for (i=0; i < ToneIDCount; i++) {
            if (((UInt32 *)ToneIDList)[i] == tone) {
                return PhnLibGetToneName(refNum, i, name, maxLength);
            }
        }
        break;
    case hsAttrPhoneTypeCDMA:
        /* CDMA version uses a slot */
        /* lame lame lame -- CDMA reboots if toneslot is invalid */
        /* so I have to validate it before I try it */
        for (i=0; i < ToneIDCount; i++) {
            if (((struct st300toneid *)ToneIDList)[i].slot == tone) {
                return Treo300PhnLibGetToneName(refNum, tone, name, maxLength);
            }
        }
        break;
    }
    return 1;
}

Err TreoGetToneNameIdx(UInt16 refNum, UInt16 toneidx, char *name,
    UInt16 maxLength)
{
    if (!PhnLibModulePowered(refNum)) return 1; /* sanity */
    switch (WhichPhoneType()) {
    case hsAttrPhoneTypeGSM:
        return PhnLibGetToneName(refNum, toneidx, name, maxLength);
    case hsAttrPhoneTypeCDMA:
        /* CDMA version uses a slot */
        return Treo300PhnLibGetToneName(refNum, toneidx, name, maxLength);
    }
    return 1;
}

Err TreoPlayTone(UInt16 refNum, UInt32 tone, UInt16 volume, Boolean vibrate,
    UInt16 duration)
{
    Err err=0;
    UInt16 tui16;
    UInt32 tui32;

    if (!PhnLibModulePowered(refNum)) return 1; /* sanity */

//    if (!HsAttrGet(hsAttrLidOpen, 0, &tui16)) {
//        /* lid open, soft volume */
//        if (tui16 && volume) volume = phnRingerSoft;
//    }

    if (!PhnLibGetPhoneCallStatus(refNum, &tui32) &&
        (tui32&(phnVoiceCall1Active|phnVoiceCall2Active))) {
        /* if you're on the phone, no volume */
        volume = 0;
    }

    if (!HsAttrGet(hsAttrRingSwitch, 0, &tui16) && !tui16) {
        /* ring switch set to mute? */
        volume = 0;
    }

    if (volume) {
        switch (WhichPhoneType()) {
        case hsAttrPhoneTypeGSM:
            switch (volume) {
            case 1: volume = phnRingerSoft; break;
            case 2: volume = 4; break; /* is there a medium for GSM? */
            case 3: volume = phnRingerLoud; break;
            }
            /* don't know how to handle duration yet */
            err = PhnLibPlayTone(refNum, tone, volume);
            break;
        case hsAttrPhoneTypeCDMA:
            /* Treo 300 PhnRingerVolumeType enum is different */
            switch (volume) {
            case 1: volume = 2; break; /* soft */
            case 2: volume = 3; break; /* medium */
            case 3: volume = 8; break; /* loud */
            }
            tui16 = tone;
            err = Treo300PhnLibPlayTone(refNum, tui16, volume, phnVibrateOff,
                2 /* phnSndTypeRinger */, duration /* duration */,
                true /* useSoundOffFlag */);
            /*
             * Normally the tone stops after duration by itself.  However,
             * if you have loud volume it will leave the amp on unless you
             * do a PhnLibStopTone().  So I do that on a timer.
             */
            FtrSet(MainAppID, FTR_STOPTONE, TimGetSeconds()+duration+2);
            set_next_alarm_ref();
            break;
        }
    }

    if (!err && vibrate) {
        /* use HsIndicator to run a vibrate sequence */
        if (!HsIndicator(hsIndicatorCmdSetSequence, 0, (void *)ics)) {
            HsIndicator(hsIndicatorCmdEnable, 0, (void *)1);
        }
    }

    return err;
}

void ToneListDrawItem(Int16 itemNum, RectanglePtr bounds, char **unusedP)
{
    Int16 tl, tw;
    UInt16 i;
    Boolean ignored;
    char ellipsisChar=chrEllipsis;
    char name[32];

    if (ToneIDList && (itemNum >= 0) && (itemNum < ToneIDCount)) {

        /*
         * The GSM expects an index, and the CDMA expects a slot which
         * may be non-linear.
         */
        i = itemNum;
        if (WhichPhoneType() == hsAttrPhoneTypeCDMA)
            i = ((struct st300toneid *)ToneIDList)[itemNum].slot;

        if (!TreoGetToneNameIdx(phonelibref, i, name, sizeof(name))) {
            tl = StrLen(name);
            tw = FntCharsWidth(name, tl);
            if (tw <= bounds->extent.x) {
                WinDrawChars(name, tl, bounds->topLeft.x,
                    bounds->topLeft.y);
            } else {
                tw = bounds->extent.x-FntCharWidth(ellipsisChar);
                FntCharsInWidth(name, &tw, &tl, &ignored);
                WinDrawChars(name, tl, bounds->topLeft.x,
                    bounds->topLeft.y);
                WinDrawChars(&ellipsisChar, 1, bounds->topLeft.x+tw,
                    bounds->topLeft.y);
            }
        } else {
            StrPrintF(name, "%d", itemNum);
            tl = StrLen(name);
            WinDrawChars(name, tl, bounds->topLeft.x,
                bounds->topLeft.y);
        }
    }
}

UInt32 ToneList(UInt16 listid, UInt16 triggerid, UInt32 curtone)
{
    Int16 curSelection=-1;
    Int16 newSelection;
    UInt16 i;
    UInt32 newtone=curtone;
    ListPtr lst;
    ControlPtr trig;
    FormPtr frm;
    CharPtr label;
    Boolean f=false;

    if (!ToneIDCount) return curtone;

    switch (WhichPhoneType()) {
    case hsAttrPhoneTypeGSM:
        /* for GSM versions, find the tone's index using toneid */
        for (i=0; i < ToneIDCount; i++) {
            if (((UInt32 *)ToneIDList)[i] == curtone) {
                f = true;
                break;
            }
        }
        break;
    case hsAttrPhoneTypeCDMA:
        /* for CDMA versions, find the tone's index using toneslot */
        for (i=0; i < ToneIDCount; i++) {
            if (((struct st300toneid *)ToneIDList)[i].slot == curtone) {
                f = true;
                break;
            }
        }
        break;
    default: i = 0; break;
    }

    if (f) curSelection = i;
    
    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listid));
    LstSetDrawFunction(lst, (ListDrawDataFuncPtr)ToneListDrawItem);
    LstSetListChoices(lst, NULL, ToneIDCount);
    LstSetSelection(lst, curSelection);
    if (curSelection >= 0) LstSetTopItem(lst, curSelection);
    newSelection = LstPopupList(lst);

    if ((newSelection != curSelection) && (newSelection >= 0)) {
        switch (WhichPhoneType()) {
        case hsAttrPhoneTypeGSM:
            /* for GSM versions, return the toneid */
            newtone = ((UInt32 *)ToneIDList)[newSelection];
            break;
        case hsAttrPhoneTypeCDMA:
            /* for CDMA versions, return the toneslot */
            newtone = ((struct st300toneid *)ToneIDList)[newSelection].slot;
            break;
        }
        /* copy the name into the popup trigger label */
        trig = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, triggerid));
        label = (char *)CtlGetLabel(trig);
        /* the 32 is how many chars we have alloc'd in the trigger */
        TreoGetToneName(phonelibref, newtone, label, 32);
        CtlSetLabel(trig, label);
    }

    return newtone;
}

Boolean is_lid_open(void)
{
    UInt16 tui16;
    if (!HsAttrGet(hsAttrLidOpen, 0, &tui16) && tui16) return true;
    return false;
}

Boolean is_display_on(void)
{
    UInt16 tui16;
    if (!HsAttrGet(hsAttrDisplayOn, 0, &tui16) && tui16) return true;
    return false;
}

Boolean was_lid_opened(void)
{
    #define MAX_KEY_PEEKS 2
    SysEventType e[MAX_KEY_PEEKS];
    Int16 i=0, j;
    Boolean islid=false;

    /*
     * Lid being opened is indicated by a hsChrLidStatusChg char.  However it's
     * not the first char enqueued, so I have to pop off all chars and then push
     * them back in order to look for it.  The alternative would be to install
     * a hack and watch the char go by.  This is safer IMO.  There's only ~3
     * chars enqueued at power on anyhow.
     *
     * The system may not actually care of the ordering, but since there are
     * very few chars, I store them and put them back into the queue in order.
     */

    /* take them out of the queue and inspect */
    while ((i < MAX_KEY_PEEKS) && !EvtDequeueKeyEvent(&e[i], false)) {
        if ((e[i].data.keyDown.chr == hsChrLidStatusChg)) {
            islid = true;
            i++;
            break;
        }
        i++;
    }
    /* put them back in */
    for (j=0; j < i; j++) {
        /*
         * I use EvtEnqueueKey rather than EvtAddEventToQueue because they
         * are treated differently in weird ways.
         */
        EvtEnqueueKey(e[j].data.keyDown.chr, e[j].data.keyDown.keyCode,
            e[j].data.keyDown.modifiers);
    }

    return islid;
}

Err dial_phone(char *txt)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType SearchState;
    Err err;
    UInt8 *p;
    PhoneAppLaunchCmdDialPtr dial;
    char *number;
    char *name;
    const char whotxt[]="Handsfree";

    err = DmGetNextDatabaseByTypeCreator(true, &SearchState,
        sysFileTApplication, hsFileCPhone, true, &cardNo, &dbID);
    if (err) return err;

    /* I need to pass some app independent memory to Phone */
    p = MemPtrNew(sizeof(PhoneAppLaunchCmdDialType)+MAX_PHONE_NUMBER+1
        +StrLen(whotxt)+1);
    if (!p) return sysErrNoFreeRAM;
    /* zero it out */
    MemSet(p, sizeof(PhoneAppLaunchCmdDialType)+MAX_PHONE_NUMBER+1
        +StrLen(whotxt)+1, 0);
    dial = (PhoneAppLaunchCmdDialPtr)p;
    number = p+sizeof(PhoneAppLaunchCmdDialType);
    name = number+MAX_PHONE_NUMBER+1;
    MemMove(number, txt, StrLen(txt));
    MemMove(name, whotxt, StrLen(whotxt));
    dial->number = number;
    dial->name = name;
    /*
     * Now I have a memory block with a PhoneAppLaunchCmdDialType structure
     * and an asciiz phone number after it, and after that our fake name.
     */
    dial->version = 1;
    dial->dialMethod = PhoneAppDialMethodNormal;
    MemPtrSetOwner(p, 0); /* owned by system */

    err = SysUIAppSwitch(cardNo, dbID, phoneAppLaunchCmdDial, p);
    return err;
}

/*
 * I put these here because I didn't want my PilotMain() to use up all that
 * stack space for each and every launchcode out there.
 */
UInt32 HandleLaunchCodes(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    PhnEventPtr pet;
    UInt16 tui16;
    UInt32 tui32;
    struct sprefs *tprefs;
    UInt16 tplibref, tnmlibref;
    SysNotifyParamType *snpt;
    SleepEventParamType *sept;
    UInt16 et;
    UInt32 retval=0;
    NetMasterAutoLoginType nmal;
    UInt16 i;
    char *p;
//    char txt[64];

//    MemPtr startPP, endPP;
//    SysGetStackInfo(&startPP, &endPP);
//    ErrFatalDisplayIf((endPP-startPP) < 4000, "HandleLaunchCodes stack < 4000");

    /* load the bare minimum prefs */
    if (!LoadTempPrefs(&tprefs, WHICHPREFS_PHONEFLAGS)) return retval;

    switch (cmd) {
    case sysAppLaunchCmdNotify:
        LoadPrefs(tprefs, WHICHPREFS_ALL); /* need more prefs for this event */
        snpt = (SysNotifyParamType *)cmdPBP;
        switch (snpt->notifyType) {
        case sysNotifySleepRequestEvent:
            sept = (SleepEventParamType *)(snpt->notifyDetailsP);
            if ((sept->reason == sysSleepPowerButton) &&
                !HsAttrGet(hsAttrSysSleepReason, 0, &tui16) &&
                (tui16 == hsSysSleepLid)) {
                if (LoadPhoneLibrary(&tplibref)) break;
                if ((tprefs->phoneflags&PHONEFLAGS_STAYONLIDNETWORK) &&
                    !PhnLibGetPhoneCallStatus(tplibref, &tui32) &&
                    (tui32&phnGPRSCallActive)) {
                    /*
                     * So, if a SleepLid request comes in, and we have
                     * our prefs set for it, and the phonecallstatus
                     * shows a data call active, then we defer sleep.
                     */
                    sept->deferSleep++;
                }
                /* make the tone even if deferred -- I like it better */
                if (tprefs->phoneflags&PHONEFLAGS_ENABLEPOWEROFFTONE) {
                    /* play the power off tone we have assigned */
                    TreoPlayTone(tplibref, tprefs->funtone[1] /* power off */,
                        tprefs->funvolume[1], false,
                        tprefs->funduration[1]);
                }
                ClosePhoneLibrary(tplibref);
            }
            break;
        case sysNotifySleepNotifyEvent:
            /* a timer already active, leave it */
            if (!FtrGet(MainAppID, FTR_DELAYLAUNCH, &tui32) && tui32) break;
            for (i=0; i < SMSENTRIES; i++) {
                /*
                 * I used to check _DELAYLAUNCH setting as well, but now
                 * that I cause a delay launch for SMS received during a
                 * phone call (see MatchSMStext), I simply check for a memory
                 * alloc in a SMS slot.
                 */
                /* see if there is feature memory alloc'd in any SMS slot */
                if (!FtrGet(MainAppID, FTR_SMS1+i, (UInt32 *)&p)) {
                    /* yup, so now we set a 5 second timer */
                    FtrSet(MainAppID, FTR_DELAYLAUNCH, TimGetSeconds()+5);
                    set_next_alarm_ref();
                    break;
                }
            }
            break;
        case sysNotifyEarlyWakeupEvent:
            if ((tprefs->phoneflags&PHONEFLAGS_ENABLEPOWERONTONE) &&
                was_lid_opened()) {
                /*
                 * Can only tell if lid was opened in early wakeup.  But
                 * some devices have flaky sound at this time.
                 */
                /* set this feature for late wakeup */
                FtrSet(MainAppID, FTR_LIDOPENED, 1);
            }
            break;
        case sysNotifyLateWakeupEvent:
            if ((tprefs->phoneflags&PHONEFLAGS_ENABLEPOWERONTONE) &&
                !FtrGet(MainAppID, FTR_LIDOPENED, &tui32) && tui32) {
                FtrSet(MainAppID, FTR_LIDOPENED, 0); /* clear it */
                if (LoadPhoneLibrary(&tplibref)) break;
                TreoPlayTone(tplibref, tprefs->funtone[0] /* power on */,
                   tprefs->funvolume[0], false,
                   tprefs->funduration[0]);
                ClosePhoneLibrary(tplibref);
            }
            break;
        case sysNotifySyncStartEvent:
            /* unregister for SMS phone events - prevents callback crashes */
            if (LoadPhoneLibrary(&tplibref)) break;
            phone_events_unregister(tplibref);
            ClosePhoneLibrary(tplibref);
            break;
        case sysNotifySyncFinishEvent:
            /* register for SMS phone events  */
            if (LoadPhoneLibrary(&tplibref)) break;
            phone_events_register(tplibref, tprefs);
            ClosePhoneLibrary(tplibref);
            break;
        }
        break;

    case phnLibLaunchCmdEvent: /* for GSM devices */
    case 0x7000: /* for Treo 300 */
    case 0xabad: /* Treo 600 */
        pet = (PhnEventPtr)cmdPBP;
        switch (pet->eventType) {
        case phnEvtPower:
            if (pet->data.power.state == phnPowerOff) {
                /* kill this timer whenever the module power turns off */
                FtrSet(MainAppID, FTR_COVERAGELOSS, 0);
                /* kill our feature for auto-login, so it'll try again */
                FtrSet(MainAppID, FTR_AUTOLOGIN, 0);
                set_next_alarm_ref();
            }
            break;
        case phnEvtRegistration:
            LoadPrefs(tprefs, WHICHPREFS_DROPNET); /* need more prefs */
            switch (pet->data.registration.status) {
            case registrationNone:
            case registrationDenied:
            case registrationUnknown:
            case registrationSearch:
                if (tprefs->phoneofftime > 0) {
                    /*
                     * Phone is in network search mode.  If our pref is set
                     * then we'll set a timer, which when it expires, we will
                     * turn off the phone.  This is an option because on some
                     * phones, the power drain is severe when doing a network
                     * search for an extended period.
                     */
                    FtrSet(MainAppID, FTR_COVERAGELOSS,
                        TimGetSeconds()+(tprefs->phoneofftime*60));
                    set_next_alarm_ref();
                }
                /*
                 * An option for dropping the network (only the 3G) when
                 * coverage is lost.  Actually this is turning out to be
                 * a significant option in our battle for super quick drain
                 * work-arounds.  At least so far so good, this has been
                 * the only option to make a difference.
                 */
                if (tprefs->dropnetcoverage > 0) {
                    FtrSet(MainAppID, FTR_DROPNETCOVERAGE,
                        TimGetSeconds()+tprefs->dropnetcoverage);
                    set_next_alarm_ref();
                }
                break;
            case registrationHome:
            case registrationRoaming:
                /*
                 * When it's registered, cancel our phoneoff
                 * alarm, if any.
                 */
                FtrSet(MainAppID, FTR_COVERAGELOSS, 0);
                FtrSet(MainAppID, FTR_DROPNETCOVERAGE, 0);
                /* also cancel "phone on" attempts */
                FtrSet(MainAppID, FTR_PHONEONEVERY, 0); /* interval timer */
                FtrSet(MainAppID, FTR_PHONEONTRIES, 0); /* clear try count */
                /*
                 * GSM phones don't auto-login to the network.  Well, there
                 * is a setting in Preferences->Network->Options for it, but
                 * it doesn't actually do a "Service Connection Progress".
                 *
                 * So I am looking at the pref and I am going to do a network
                 * connection when we come out of registration search.  I'll
                 * only do this once per power-on -- that's what my feature
                 * is for.
                 */
                if ((WhichPhoneType() == hsAttrPhoneTypeGSM) &&
                    (tprefs->phoneflags&PHONEFLAGS_AUTOLOGIN) &&
                    !UseNetMasterLibrary(&tnmlibref) &&
                    !NetMasterAutoLoginSettingGet(tnmlibref, &nmal) &&
                    (nmal&netMasterAutoLoginWhenWirelessOn)) {
                    /* only do it once per phone cycle */
                    if (FtrGet(MainAppID, FTR_AUTOLOGIN, &tui32) || (!tui32)) {
                        /* ftr to auto-login in 10 sec, only needed for GSM */
                        FtrSet(MainAppID, FTR_AUTOLOGIN, TimGetSeconds()+10);
                    }
                }
                set_next_alarm_ref();
                break;
            default: break;
            }
            break;

        case phnEvtKeyPress:
            /* testing */
            //tui32 = pet->data.keyPressed.key;
            //StrPrintF(txt, "key %08lx", tui32);
            //FrmCustomAlert(ErrorAlert, txt, " ", " ");

            if (!(tprefs->phoneflags&PHONEFLAGS_ENABLEHANDSFREEDIAL)) break;
            if (LoadPhoneLibrary(&tplibref)) break;
            /*
             * Don't do anything if a call is in progress.  That's because
             * the button is used to hangup.
             */
            if (!PhnLibGetPhoneCallStatus(tplibref, &tui32) &&
                !(tui32&(phnVoiceCall1Active|phnVoiceCall2Active))) {
                LoadPrefs(tprefs, WHICHPREFS_ALL); /* need more prefs */
                /* pre OS 5.x had different button values for CDMA */
                if (WhichPhoneType() == hsAttrPhoneTypeCDMA) {
                    /* these are completely different than GSM */
                    tui16 = 3; /* value for Treo 300 */
                    if (is_os_5x()) tui16 = 0; /* value for Treo 600 */
                    if ((pet->data.keyPressed.key == tui16) &&
                        (pet->data.keyPressed.modifiers&1)) {
                        /* handsfree button is pressed */
                        /* schedule a timer to dial a number, can't do it now */
                        FtrSet(MainAppID, FTR_HANDSFREEDIAL, TimGetSeconds());
                        set_next_alarm_ref();
                    }
                } else {
                    tui16 = 2; /* value for Treo 270 */
                    if (is_os_5x()) tui16 = 0; /* value for Treo 600 */
                    if ((pet->data.keyPressed.key == tui16) &&
                        ((pet->data.keyPressed.modifiers&1) ||
                        (pet->data.keyPressed.modifiers&phnButtonHeld))) {
                        /* handsfree button is pressed */
                        /* schedule a timer to dial a number, can't do it now */
                        FtrSet(MainAppID, FTR_HANDSFREEDIAL, TimGetSeconds());
                        set_next_alarm_ref();
                    }
                }
            }
            ClosePhoneLibrary(tplibref);
            break;

        case 29: /* phnEvtForceUpdateTitleBar - Treo 300 only */
            /*
             * This event is nifty.  Not aware of an equivalent for GSM
             * devices.  It fires whenever there is a change in signal
             * quality - typical every few seconds.
             */
            break;

        case 32: /* phnEvtOneXStatus - Treo 300 only */
            if (WhichPhoneType() != hsAttrPhoneTypeCDMA) break;
            LoadPrefs(tprefs, WHICHPREFS_DROPNET); /* need more prefs */
            set_dropnet_times(tprefs, 0);
            break;

        default: /* default case */
        /*
         * The enum is different on GSM and CDMA so I'm going to have
         * to fudge the eventType, and make another switch.  The CDMA enum
         * is missing phnEvtPassword and so every value after phnEvtPower is
         * -1 on the Treo 300.
         */
        et = pet->eventType;
        /* CDMA earlier than 5.x had a broken event enum */
        if ((WhichPhoneType() == hsAttrPhoneTypeCDMA) && !is_os_5x()) et++;
        switch (et) {
        case phnEvtDisconnectInd: /* this occurs with an error dialog */
            /*
             * You know the "Connection Progress" dialog.  Well, the stupid
             * thing will sit there waiting for an OK if it encounters an
             * error.  This causes problems with scheduled networking apps
             * because it effectively blocks the whole system.  So we
             * set up a timer which uses some nifty Handspring NetMaster
             * functions to clear this dialog.
             *
             * Note that I chose phnEvtDisconnectInd because I found that it
             * fired when I needed it to.  I had a hard time finding an
             * event which happens due to the dialog popping up.
             *
             * This event also occurs in some other circumstances, such as
             * when hanging up a phone call.
             *
             * I couldn't find an event specifically for when the "Connection
             * Progress" dialog or error dialog started, so I just put in a
             * little beep and tested it on a 300 device and found one which
             * was close enough.
             *
             * I then initiate a timer and in the timer I check whether the
             * dialog is present with the NetMaster calls.
             */
            if (tprefs->phoneflags&PHONEFLAGS_AUTOOKCONNECT) {
                FtrSet(MainAppID, FTR_AUTOOKCONNECT, TimGetSeconds()+5);
                set_next_alarm_ref();
            }
            break;

        case phnEvtMessageInd: /* SMS */
        case phnEvtSegmentInd: /* I've been told to do this for Sprint 650 */
            if (HandleSMS(pet->data.params.id)) pet->acknowledge = true;
            break;

        }
        break; /* default case */
        }
        break;

    case sysAppLaunchCmdCustomBase+1: /* stop playing a tone */
        /*
         * This is not super well designed, but when I launch something,
         * such as PBJTime, I need some way to have user activity stop the
         * currently playing tone.  Rather than adding this PhoneLib stuff
         * to PBJTime, I've made this launchcode.
         */
        if (!LoadPhoneLibrary(&tplibref)) {
            PhnLibStopTone(tplibref);
            ClosePhoneLibrary(tplibref);
        }
        break;

    }

    UnloadTempPrefs(tprefs);
    return retval;
}

/*
 * The program starts here.
 */
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    Err err;
    struct sprefs *tprefs;

    switch (cmd) {
    case sysAppLaunchCmdNormalLaunch:
        if (!is_os_32()) {
            FrmAlert(RomIncompatibleAlert);
            return sysErrRomIncompatible;
        }
        /*
         * I guess we are an appropriate ROM version so let's run.
         */
        err = StartApplication();
        if (err) return err;
        EventLoop();
        StopApplication();
        break;

    case sysAppLaunchCmdSystemReset:
        /*
         * Give it 10 seconds to settle down & stuff.  This is just a cosmetic
         * issue, as I haven't had any problems with turning on phone power
         * immediately, but it looks better if we wait until the Palm has
         * completed booting.
         */
        FtrSet(MainAppID, FTR_RESET, TimGetSeconds()+10);
        set_next_alarm_ref();
        register_notifyevent(sysNotifySleepRequestEvent);
        register_notifyevent(sysNotifySleepNotifyEvent);
        register_notifyevent(sysNotifyEarlyWakeupEvent);
        register_notifyevent(sysNotifyLateWakeupEvent);
        /* I register for this so I can unregister for PhoneLib events */
        register_notifyevent(sysNotifySyncStartEvent);
        register_notifyevent(sysNotifySyncFinishEvent);
        /* no break wanted here */
    //case sysAppLaunchCmdSyncNotify:
    case sysAppLaunchCmdTimeChange:
        if (LoadTempPrefs(&tprefs, WHICHPREFS_ALL)) {
            set_phoneonoff_times(tprefs, 0); /* init phone times */
            set_dropnet_times(tprefs, 0); /* init dropnet times */
            UnloadTempPrefs(tprefs);
        }
        break;

    case phnLibLaunchCmdEvent: /* for GSM devices */
    case 0x7000: /* for Treo 300 */
    case 0xabad: /* Treo 600 */
    case sysAppLaunchCmdNotify:
    case sysAppLaunchCmdCustomBase+1:
        return HandleLaunchCodes(cmd, cmdPBP, launchFlags);

    case sysAppLaunchCmdAlarmTriggered:
        AlarmTriggered((SysAlarmTriggeredParamType *)cmdPBP);
        break;
    case sysAppLaunchCmdDisplayAlarm:
        DisplayAlarm((SysDisplayAlarmParamType *)cmdPBP);
        break;

    default:
        return sysErrParamErr;
    }

    return 0;
}
